############################################################################
# Copyright (c) 2025-2026 University of Helsinki
# All Rights Reserved
# See file LICENSE for details.
############################################################################

import pytest
from src.barcode_calling.indexers import KmerIndexer, Dict2BitKmerIndexer, ArrayKmerIndexer, Array2BitKmerIndexer
from src.barcode_calling.common import str_to_2bit, batch_str_to_2bit


class TestKmerIndexer:
    """Test basic dictionary-based KmerIndexer."""

    def test_init(self):
        """Test indexer initialization."""
        barcodes = ["ACTG", "TGCA", "GGGG"]
        indexer = KmerIndexer(barcodes, kmer_size=3)

        assert len(indexer.seq_list) == 3
        assert indexer.k == 3
        assert not indexer.empty()

    def test_empty(self):
        """Test empty index detection."""
        indexer = KmerIndexer([], kmer_size=3)
        assert indexer.empty()

    def test_get_kmers(self):
        """Test k-mer generation."""
        indexer = KmerIndexer([], kmer_size=3)
        kmers = list(indexer._get_kmers("ACTGAC"))

        assert len(kmers) == 4
        assert kmers == ["ACT", "CTG", "TGA", "GAC"]

    def test_get_kmers_short_sequence(self):
        """Test k-mer generation for short sequence."""
        indexer = KmerIndexer([], kmer_size=5)
        kmers = list(indexer._get_kmers("ACT"))

        assert len(kmers) == 0

    def test_exact_match(self):
        """Test finding exact match."""
        barcodes = ["ACTGACTG", "TGCATGCA", "GGGGGGGG"]
        indexer = KmerIndexer(barcodes, kmer_size=4)

        results = indexer.get_occurrences("ACTGACTG")

        assert len(results) > 0
        assert results[0][0] == "ACTGACTG"
        assert results[0][1] > 0  # Shared k-mer count

    def test_similar_match(self):
        """Test finding similar sequence."""
        barcodes = ["ACTGACTG", "TGCATGCA"]
        indexer = KmerIndexer(barcodes, kmer_size=4)

        # Query with 1 mismatch
        results = indexer.get_occurrences("ACTGACTT")

        assert len(results) > 0
        # Should find ACTGACTG as closest match
        assert results[0][0] == "ACTGACTG"

    def test_min_kmers_filter(self):
        """Test minimum k-mer threshold."""
        barcodes = ["ACTGACTG", "TGCATGCA"]
        indexer = KmerIndexer(barcodes, kmer_size=4)

        # Very different sequence
        results = indexer.get_occurrences("CCCCCCCC", min_kmers=3)

        # Should have fewer results with higher threshold
        assert len(results) <= 2

    def test_max_hits(self):
        """Test maximum hits limit."""
        barcodes = ["ACTG", "ACTC", "ACTA", "ACTT"]
        indexer = KmerIndexer(barcodes, kmer_size=3)

        results = indexer.get_occurrences("ACTG", max_hits=2)

        assert len(results) <= 2

    def test_ignore_equal(self):
        """Test ignoring exact matches."""
        barcodes = ["ACTG", "TGCA"]
        indexer = KmerIndexer(barcodes, kmer_size=3)

        results = indexer.get_occurrences("ACTG", ignore_equal=True)

        # Should not return the exact match
        for seq, _, _ in results:
            assert seq != "ACTG"

    def test_append(self):
        """Test adding barcode dynamically."""
        barcodes = ["ACTG"]
        indexer = KmerIndexer(barcodes, kmer_size=3)

        indexer.append("TGCA")

        assert len(indexer.seq_list) == 2
        results = indexer.get_occurrences("TGCA")
        assert any(seq == "TGCA" for seq, _, _ in results)

    def test_result_format(self):
        """Test result tuple format."""
        barcodes = ["ACTGACTG"]
        indexer = KmerIndexer(barcodes, kmer_size=4)

        results = indexer.get_occurrences("ACTGACTG")

        assert len(results) > 0
        # Result should be (sequence, count, positions)
        seq, count, positions = results[0]
        assert isinstance(seq, str)
        assert isinstance(count, int)
        assert isinstance(positions, list)


class TestDict2BitKmerIndexer:
    """Test memory-efficient Dict2BitKmerIndexer with 2-bit encoded sequences."""

    def test_init(self):
        """Test indexer initialization with 2-bit encoded sequences."""
        barcodes = ["ACTGACTG", "TGCATGCA", "GGGGGGGG"]
        bit_barcodes = batch_str_to_2bit(barcodes, seq_len=8)
        indexer = Dict2BitKmerIndexer(bit_barcodes, kmer_size=3, seq_len=8)

        assert len(indexer.seq_list) == 3
        assert indexer.k == 3
        assert not indexer.empty()

    def test_empty(self):
        """Test empty index detection."""
        indexer = Dict2BitKmerIndexer([], kmer_size=3, seq_len=8)
        assert indexer.empty()

    def test_exact_match(self):
        """Test finding exact match."""
        barcodes = ["ACTGACTG", "TGCATGCA", "GGGGGGGG"]
        bit_barcodes = batch_str_to_2bit(barcodes, seq_len=8)
        indexer = Dict2BitKmerIndexer(bit_barcodes, kmer_size=4, seq_len=8)

        results = indexer.get_occurrences("ACTGACTG")

        assert len(results) > 0
        assert results[0][0] == "ACTGACTG"
        assert results[0][1] > 0  # Shared k-mer count

    def test_similar_match(self):
        """Test finding similar sequence."""
        barcodes = ["ACTGACTG", "TGCATGCA"]
        bit_barcodes = batch_str_to_2bit(barcodes, seq_len=8)
        indexer = Dict2BitKmerIndexer(bit_barcodes, kmer_size=4, seq_len=8)

        # Query with 1 mismatch
        results = indexer.get_occurrences("ACTGACTT")

        assert len(results) > 0
        assert results[0][0] == "ACTGACTG"

    def test_min_kmers_filter(self):
        """Test minimum k-mer threshold."""
        barcodes = ["ACTGACTG", "TGCATGCA"]
        bit_barcodes = batch_str_to_2bit(barcodes, seq_len=8)
        indexer = Dict2BitKmerIndexer(bit_barcodes, kmer_size=4, seq_len=8)

        # Very different sequence
        results = indexer.get_occurrences("CCCCCCCC", min_kmers=3)

        assert len(results) <= 2

    def test_max_hits(self):
        """Test maximum hits limit."""
        barcodes = ["ACTGACTG", "ACTCACTC", "ACTAACTA", "ACTTACTT"]
        bit_barcodes = batch_str_to_2bit(barcodes, seq_len=8)
        indexer = Dict2BitKmerIndexer(bit_barcodes, kmer_size=3, seq_len=8)

        results = indexer.get_occurrences("ACTGACTG", max_hits=2)

        assert len(results) <= 2

    def test_ignore_equal(self):
        """Test ignoring exact matches."""
        barcodes = ["ACTGACTG", "TGCATGCA"]
        bit_barcodes = batch_str_to_2bit(barcodes, seq_len=8)
        indexer = Dict2BitKmerIndexer(bit_barcodes, kmer_size=3, seq_len=8)

        results = indexer.get_occurrences("ACTGACTG", ignore_equal=True)

        for seq, _, _ in results:
            assert seq != "ACTGACTG"

    def test_consistency_with_kmerindexer(self):
        """Test that results match basic KmerIndexer."""
        barcodes = ["ACTGACTG", "TGCATGCA", "GGGGGGGG"]
        basic = KmerIndexer(barcodes, kmer_size=4)
        bit_barcodes = batch_str_to_2bit(barcodes, seq_len=8)
        dict2bit = Dict2BitKmerIndexer(bit_barcodes, kmer_size=4, seq_len=8)

        query = "ACTGACTT"
        basic_results = basic.get_occurrences(query)
        dict2bit_results = dict2bit.get_occurrences(query)

        # Should find same sequences
        basic_seqs = set(seq for seq, _, _ in basic_results)
        dict2bit_seqs = set(seq for seq, _, _ in dict2bit_results)
        assert basic_seqs == dict2bit_seqs

    def test_consistency_with_arraykmerindexer(self):
        """Test that results match ArrayKmerIndexer."""
        barcodes = ["ACTGACTG", "TGCATGCA", "GGGGGGGG"]
        array = ArrayKmerIndexer(barcodes, kmer_size=4)
        bit_barcodes = batch_str_to_2bit(barcodes, seq_len=8)
        dict2bit = Dict2BitKmerIndexer(bit_barcodes, kmer_size=4, seq_len=8)

        query = "ACTGACTT"
        array_results = array.get_occurrences(query)
        dict2bit_results = dict2bit.get_occurrences(query)

        # Should find same sequences
        array_seqs = set(seq for seq, _, _ in array_results)
        dict2bit_seqs = set(seq for seq, _, _ in dict2bit_results)
        assert array_seqs == dict2bit_seqs

    def test_result_format(self):
        """Test result tuple format."""
        barcodes = ["ACTGACTG"]
        bit_barcodes = batch_str_to_2bit(barcodes, seq_len=8)
        indexer = Dict2BitKmerIndexer(bit_barcodes, kmer_size=4, seq_len=8)

        results = indexer.get_occurrences("ACTGACTG")

        assert len(results) > 0
        seq, count, positions = results[0]
        assert isinstance(seq, str)
        assert isinstance(count, int)
        assert isinstance(positions, list)

    def test_index_uses_int_keys(self):
        """Test that index uses integer keys, not strings."""
        barcodes = ["ACTGACTG"]
        bit_barcodes = batch_str_to_2bit(barcodes, seq_len=8)
        indexer = Dict2BitKmerIndexer(bit_barcodes, kmer_size=3, seq_len=8)

        # All keys should be integers
        for key in indexer.index.keys():
            assert isinstance(key, int)

    def test_sequences_stored_as_integers(self):
        """Test that sequences are stored as 2-bit encoded integers."""
        barcodes = ["ACTGACTG"]
        bit_barcodes = batch_str_to_2bit(barcodes, seq_len=8)
        indexer = Dict2BitKmerIndexer(bit_barcodes, kmer_size=3, seq_len=8)

        # Sequences should be integers
        for seq in indexer.seq_list:
            assert isinstance(seq, (int, type(bit_barcodes[0])))


class TestArrayKmerIndexer:
    """Test optimized array-based KmerIndexer."""

    def test_init(self):
        """Test indexer initialization."""
        barcodes = ["ACTG", "TGCA"]
        indexer = ArrayKmerIndexer(barcodes, kmer_size=3)

        assert len(indexer.seq_list) == 2
        assert indexer.k == 3
        assert len(indexer.index) == 64  # 4^3

    def test_binary_encoding(self):
        """Test nucleotide to binary conversion."""
        assert ArrayKmerIndexer.NUCL2BIN['A'] == 0
        assert ArrayKmerIndexer.NUCL2BIN['C'] == 1
        assert ArrayKmerIndexer.NUCL2BIN['G'] == 2
        assert ArrayKmerIndexer.NUCL2BIN['T'] == 3

    def test_get_kmer_indexes(self):
        """Test k-mer index generation."""
        indexer = ArrayKmerIndexer([], kmer_size=2)
        # "AC" = 00 01 = 1
        # "CT" = 01 11 = 7
        kmer_idxs = list(indexer._get_kmer_indexes("ACT"))

        assert len(kmer_idxs) == 2
        assert kmer_idxs[0] == 1  # AC
        assert kmer_idxs[1] == 7  # CT

    def test_exact_match(self):
        """Test finding exact match."""
        barcodes = ["ACTGACTG", "TGCATGCA"]
        indexer = ArrayKmerIndexer(barcodes, kmer_size=4)

        results = indexer.get_occurrences("ACTGACTG")

        assert len(results) > 0
        assert results[0][0] == "ACTGACTG"

    def test_consistency_with_basic(self):
        """Test that results match basic KmerIndexer."""
        barcodes = ["ACTGACTG", "TGCATGCA", "GGGGGGGG"]
        basic = KmerIndexer(barcodes, kmer_size=4)
        array = ArrayKmerIndexer(barcodes, kmer_size=4)

        query = "ACTGACTT"
        basic_results = basic.get_occurrences(query)
        array_results = array.get_occurrences(query)

        # Should find same sequences (order/counts might differ slightly)
        basic_seqs = set(seq for seq, _, _ in basic_results)
        array_seqs = set(seq for seq, _, _ in array_results)
        assert basic_seqs == array_seqs


class TestArray2BitKmerIndexer:
    """Test memory-efficient 2-bit KmerIndexer."""

    def test_init(self):
        """Test indexer initialization with 2-bit sequences."""
        barcodes = ["ACTGACTGACTGACTGACTGACTGA"]  # 25-mer
        bin_seqs = [str_to_2bit(b) for b in barcodes]
        indexer = Array2BitKmerIndexer(bin_seqs, kmer_size=6, seq_len=25)

        assert indexer.total_sequences == 1
        assert indexer.seq_len == 25
        assert indexer.k == 6
        assert not indexer.empty()

    def test_empty(self):
        """Test empty index detection."""
        indexer = Array2BitKmerIndexer([], kmer_size=6, seq_len=25)
        assert indexer.empty()

    def test_get_kmer_bin_indexes(self):
        """Test k-mer extraction from 2-bit sequence."""
        # Create simple sequence
        bin_seq = str_to_2bit("ACTGAA")
        indexer = Array2BitKmerIndexer([], kmer_size=3, seq_len=6)

        kmer_idxs = list(indexer._get_kmer_bin_indexes(bin_seq))

        assert len(kmer_idxs) == 4  # 6 - 3 + 1

    def test_exact_match(self):
        """Test finding exact match."""
        seq = "ACTGACTGACTGACTGACTGACTGA"
        bin_seqs = [str_to_2bit(seq)]
        indexer = Array2BitKmerIndexer(bin_seqs, kmer_size=8, seq_len=25)

        results = indexer.get_occurrences(seq)

        assert len(results) > 0
        assert results[0][0] == seq

    def test_similar_match(self):
        """Test finding similar sequence."""
        barcode1 = "ACTGACTGACTGACTGACTGACTGA"
        barcode2 = "TGCATGCATGCATGCATGCATGCAT"
        bin_seqs = [str_to_2bit(b) for b in [barcode1, barcode2]]
        indexer = Array2BitKmerIndexer(bin_seqs, kmer_size=8, seq_len=25)

        # Query with 1 mismatch from barcode1
        query = "ACTGACTGACTGACTGACTGACTGT"
        results = indexer.get_occurrences(query)

        assert len(results) > 0
        # Should find barcode1 as closest
        assert results[0][0] == barcode1

    def test_max_hits(self):
        """Test maximum hits limit."""
        barcodes = [
            "ACTGACTGACTGACTGACTGACTGA",
            "ACTGACTGACTGACTGACTGACTGC",
            "ACTGACTGACTGACTGACTGACTGG"
        ]
        bin_seqs = [str_to_2bit(b) for b in barcodes]
        indexer = Array2BitKmerIndexer(bin_seqs, kmer_size=8, seq_len=25)

        results = indexer.get_occurrences("ACTGACTGACTGACTGACTGACTGA", max_hits=2)

        assert len(results) <= 2

    def test_flat_index_structure(self):
        """Test that index uses flat array structure."""
        barcodes = ["ACTGACTGACTGACTGACTGACTGA"]
        bin_seqs = [str_to_2bit(b) for b in barcodes]
        indexer = Array2BitKmerIndexer(bin_seqs, kmer_size=6, seq_len=25)

        # Index should be flat list
        assert isinstance(indexer.index, list)
        # Index ranges should map to flat index
        assert isinstance(indexer.index_ranges, list)
        assert len(indexer.index_ranges) == 4**6 + 1  # 4^k + 1


if __name__ == '__main__':
    pytest.main([__file__, '-v'])



############################################################################
# Copyright (c) 2024-2026 University of Helsinki
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import sys
from enum import IntEnum


class IsoQuantExitCode(IntEnum):
    """Exit codes for IsoQuant."""

    # Success
    SUCCESS = 0

    # Input/Output Errors (1-19)
    INPUT_FILE_NOT_FOUND = 1
    OUTPUT_ALREADY_EXISTS = 2
    BAM_NOT_INDEXED = 3
    GENE_DB_NOT_FOUND = 4
    REFERENCE_NOT_FOUND = 5
    NO_INPUT_DATA = 6
    FILE_FORMAT_ERROR = 7
    YAML_PARSING_ERROR = 8
    INVALID_FILE_FORMAT = 9

    # Configuration Errors (20-39)
    INVALID_PARAMETER = 20
    MISSING_REQUIRED_OPTION = 21
    INCOMPATIBLE_OPTIONS = 22
    RESUME_CONFIG_NOT_FOUND = 23
    CORRUPTED_GTF = 24
    INVALID_CONVERSION_MODE = 25

    # Data Validation Errors (40-59)
    LABEL_COUNT_MISMATCH = 40
    DUPLICATE_FILES = 41
    CHROMOSOME_MISMATCH = 42
    REPLICA_COUNT_MISMATCH = 43

    # Alignment Errors (60-79)
    ALIGNER_NOT_FOUND = 60
    INDEXING_FAILED = 61
    ALIGNMENT_FAILED = 62
    SAMTOOLS_FAILED = 63
    SUBPROCESS_FAILED = 64

    # Barcode Calling Errors (80-89)
    BARCODE_FILE_COUNT_MISMATCH = 80
    BARCODE_WHITELIST_MISSING = 81

    # Test/Runtime Errors (90-99)
    TEST_FAILED = 90
    UNCAUGHT_EXCEPTION = 99

    # Standard Error (for utilities)
    STANDARD_ERROR = 1


def exit_with_code(code: IsoQuantExitCode, message: str = None):
    """
    Exit with a specific error code and optional message.

    Parameters
    ----------
    code : IsoQuantExitCode
        The exit code to use
    message : str, optional
        Additional error message to log
    """
    if message:
        import logging
        logger = logging.getLogger('IsoQuant')
        if code == IsoQuantExitCode.SUCCESS:
            logger.info(message)
        else:
            logger.error(message)
    sys.exit(code)




############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# All Rights Reserved
# See file LICENSE for details.
############################################################################

import json
import sys
import os


def simplify_and_sum_transcripts(data):
    gene_totals_across_conditions = {}
    simplified_data = {}

    # Sum transcript values and collect them across all conditions
    for sample_id, genes in data.items():
        simplified_data[sample_id] = {}
        for gene_id, gene_data in genes.items():
            transcripts = gene_data.get("transcripts", {})
            total_value = 0.0
            simplified_transcripts = {}
            for transcript_id, transcript_details in transcripts.items():
                transcript_value = (
                    transcript_details.get("value", 0.0)
                    if isinstance(transcript_details, dict)
                    else 0.0
                )
                simplified_transcripts[transcript_id] = transcript_value
                total_value += transcript_value

            gene_data_copy = (
                gene_data.copy()
            )  # Make a copy to avoid modifying the original
            gene_data_copy["transcripts"] = simplified_transcripts
            gene_data_copy["value"] = (
                total_value  # Replace the gene-level value with the sum of transcript values
            )
            simplified_data[sample_id][gene_id] = gene_data_copy

            if gene_id not in gene_totals_across_conditions:
                gene_totals_across_conditions[gene_id] = []
            gene_totals_across_conditions[gene_id].append(total_value)

    # Determine which genes to remove
    genes_to_remove = [
        gene_id
        for gene_id, totals in gene_totals_across_conditions.items()
        if all(total < 5 for total in totals)
    ]

    # Remove genes from the simplified data structure
    for sample_id, genes in simplified_data.items():
        for gene_id in genes_to_remove:
            if gene_id in genes:
                del genes[gene_id]

    return simplified_data


def read_json(file_path):
    with open(file_path, "r") as file:
        return json.load(file)


def write_json(data, file_path):
    with open(file_path, "w") as file:
        json.dump(data, file, indent=4)


def main():
    if len(sys.argv) != 2:
        print("Usage: python script.py <input_file_path>")
        sys.exit(1)

    input_file_path = sys.argv[1]
    base, ext = os.path.splitext(input_file_path)
    output_file_path = f"{base}_simplified{ext}"

    try:
        # Load the gene data from the specified input JSON file
        gene_dict = read_json(input_file_path)

        # Simplify the transcripts, sum their values, and remove genes under a threshold across all conditions
        modified_gene_dict = simplify_and_sum_transcripts(gene_dict)

        # Save the modified gene data to the newly named output JSON file
        write_json(modified_gene_dict, output_file_path)

        print(f"Modified gene data has been saved to {output_file_path}")

    except Exception as e:
        print(f"Error: {str(e)}")
        sys.exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# Copyright (c) 2019-2022 Saint Petersburg State University
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import json
import logging
import os
import sys
import gffutils
import argparse
from traceback import print_exc
import gzip

from .error_codes import IsoQuantExitCode

logger = logging.getLogger('IsoQuant')


def db2gtf(db, gtf, _=None):
    logger.info("Converting gene annotation file to .gtf format (takes a while)...")
    with open(gtf, "w") as f:
        gene_db = gffutils.FeatureDB(db)
        for g in gene_db.features_of_type('gene', order_by=('seqid', 'start')):
            f.write(str(g) + '\n')
            for t in gene_db.children(g, featuretype=('transcript', 'mRNA')):
                f.write(str(t) + '\n')
                for e in gene_db.children(t, featuretype=('exon', 'CDS')):
                    f.write(str(e) + '\n')
    logger.info("Gene database written to " + gtf)


def db2bed(db, bed, _=None):
    # adapted from paftools in mimimap2
    colors = {
        'protein_coding': '0,128,255',
        'mRNA': '0,128,255',
        'lincRNA': '0,192,0',
        'snRNA': '0,192,0',
        'miRNA': '0,192,0',
        'misc_RNA': '0,192,0'
    }

    def get_color(transcript_kind):
        if transcript_kind in colors:
            return colors[transcript_kind]
        return '196,196,196'

    logger.info("Converting gene annotation file %s to .bed format" % db)
    genedb = gffutils.FeatureDB(db, keep_order=True)
    with open(bed, "w") as f:
        for record in genedb.all_features(featuretype=('transcript', 'mRNA')):
            if "transcript_type" in record.attributes:
                transcript_type = record["transcript_type"][0]
            else:
                transcript_type = ""
            if "gene_name" in record.attributes:
                gene_name = record["gene_name"][0]
            elif "gene_id" in record.attributes:
                gene_name = record["gene_id"][0]
            elif "Parent" in record.attributes:
                gene_name = record["Parent"][0]
            else:
                gene_name = "unknown_gene"
            transcript_name = record.id + "|" + transcript_type + "|" + gene_name
            exons = []
            cds = []
            for e in genedb.children(record, order_by='start', featuretype='exon'):
                exons.append((e.start - 1, e.end))
            for e in genedb.children(record, order_by='start', featuretype='CDS'):
                cds.append((e.start - 1, e.end))

            if not exons:
                continue
            transcript_start = exons[0][0]
            transcript_end = exons[-1][1]
            exons_lens = []
            exon_start = []
            for e in exons:
                exons_lens.append(e[1] - e[0])
                exon_start.append(e[0] - transcript_start)

            if cds:
                thick_start, thick_end = cds[0][0], cds[-1][1]
            else:
                thick_start, thick_end = transcript_start, transcript_end
            exon_lengths = ",".join([str(x) for x in exons_lens]) + ","
            exon_starts = ",".join([str(x) for x in exon_start]) + ","
            line = "%s\t%d\t%d\t%s\t1000\t%s\t%d\t%d\t%s\t%d\t%s\t%s\n" % \
                   (record.seqid, transcript_start, transcript_end, transcript_name, record.strand,
                    thick_start, thick_end, get_color(transcript_type), len(exons_lens), exon_lengths, exon_starts)
            f.write(line)
    logger.info("Gene database BED written to " + bed)


def check_input_gtf(gtf, db, complete_db):
    logger.info("Checking input gene annotation")
    gtf_is_correct, corrected_gtf, out_fname, has_meta_features = check_gtf_duplicates(gtf)
    if not gtf_is_correct:
        outdir = os.path.dirname(db)
        logger.error("Input GTF seems to be corrupted (see warnings above).")
        if out_fname and corrected_gtf:
            new_gtf_path = os.path.join(outdir, out_fname)
            with open(new_gtf_path, "w") as out_gtf:
                out_gtf.write(corrected_gtf)
            logger.error("An attempt to correct this GTF was made, the result is written to %s" % new_gtf_path)
            logger.error("NB! some transcript / gene ids in the corrected annotation are modified.")
        logger.error("Provide a correct GTF by fixing the original input GTF or checking the corrected one.")
        sys.exit(IsoQuantExitCode.CORRUPTED_GTF)
    else:
        logger.info("Gene annotation seems to be correct")

    if complete_db:
        if has_meta_features == -1:
            logger.error("You set --complete_genedb option but the provided annotation "
                         "does not contain any gene or transcript records.")
            logger.error("The results of this run will not be meaningful.")
            logger.error("Remove the output folder and restart IsoQuant without --complete_genedb.")
            sys.exit(IsoQuantExitCode.CORRUPTED_GTF)
        elif has_meta_features == 0:
            logger.warning("You set --complete_genedb option but it looks like the provided annotation "
                           "does not contain all necessary gene or transcript records.")
            logger.warning("The results of this run might not be correct.")
            logger.warning("Remove the output folder and restart IsoQuant without --complete_genedb.")


def gtf2db(gtf, db, complete_db=False, check_gtf=True):
    if check_gtf:
        check_input_gtf(gtf, db, complete_db)

    logger.info("Converting gene annotation file to .db format (takes a while)...")
    gffutils.create_db(gtf, db, force=True, keep_order=True, merge_strategy='error',
                       sort_attribute_values=True, disable_infer_transcripts=complete_db,
                       disable_infer_genes=complete_db)
    logger.info("Gene database written to " + db)
    logger.info("Provide this database next time to avoid excessive conversion")


def convert_gtf_to_db(args):
    gtf_filename = args.genedb
    gtf_filename = os.path.abspath(gtf_filename)
    genedb_filename = args.genedb_filename
    gtf_filename, genedb_filename = convert_db(gtf_filename, genedb_filename, gtf2db, args)
    return genedb_filename


def convert_db_to_gtf(args):
    genedb_filename = os.path.abspath(args.genedb)
    gtf_filename = os.path.join(args.output, os.path.splitext(os.path.basename(genedb_filename))[0] + ".gtf")
    gtf_filename, genedb_filename = convert_db(gtf_filename, genedb_filename, db2gtf, args)
    return gtf_filename


def check_gtf_duplicates(gtf):
    gtf_correct = True
    line_count = 0
    gene_ids = {}
    transcript_ids = {}
    exon_gene_ids = set()
    exon_transcript_ids = set()
    corrected_gtf = ""

    gtf_name = os.path.basename(gtf)
    gtf_name, outer_ext = os.path.splitext(gtf_name)
    if outer_ext.lower() in ['.gz', '.gzip', '.bgz']:
        handle = gzip.open(gtf, "rt")
        gtf_name, inner_ext = os.path.splitext(gtf_name)
    else:
        handle = open(gtf, "rt")
        inner_ext = outer_ext

    if inner_ext.lower() == 'gff3':
        return check_gff3_duplicates(handle)

    gff3_checked = False
    for l in handle.readlines():
        line_count += 1
        if l.startswith("#"):
            corrected_gtf += l
            continue
        v = l.strip().split("\t")
        if len(v) < 9:
            corrected_gtf += l
            continue

        attribute_column = v[8]
        if not gff3_checked:
            gff3_checked = True
            if attribute_column.find("ID=") != -1:
                handle.seek(0)
                return check_gff3_duplicates(handle)

        attrs = attribute_column.split(" ")

        gene_id_pos = -1
        for i in range(len(attrs)):
            if attrs[i] == 'gene_id':
                gene_id_pos = i
        if gene_id_pos in [-1, len(attrs) - 1]:
            logger.warning("Malformed GTF line %d (gene_id attribute value cannot be found)" % line_count)
            logger.warning(l.strip())
            gtf_correct = False
            continue

        feature_type = v[2]
        gene_str = attrs[gene_id_pos + 1]
        start_pos = gene_str.find('"')
        end_pos = gene_str.rfind('"')
        gene_id = gene_str[start_pos+1:end_pos]
        if feature_type == "gene":
            if gene_id in gene_ids:
                logger.warning("Duplicated gene id %s on line %d" % (gene_id, line_count))
                gtf_correct = False
                gene_ids[gene_id] += 1
                gene_id += ".%d" % gene_ids[gene_id]
            else:
                gene_ids[gene_id] = 0
        elif gene_id in gene_ids and gene_ids[gene_id] > 0:
            gene_id += ".%d" % gene_ids[gene_id]

        transcript_id_pos = -1
        for i in range(len(attrs)):
            if attrs[i] == 'transcript_id':
                transcript_id_pos = i
        if feature_type != "gene" and transcript_id_pos in [-1, len(attrs) - 1]:
            logger.warning("Malformed GTF line %d (transcript_id attribute value cannot be found)" % line_count)
            logger.warning(l.strip())
            gtf_correct = False
            continue

        transcript_str = attrs[transcript_id_pos + 1]
        start_pos = transcript_str.find('"')
        end_pos = transcript_str.rfind('"')
        transcript_id = transcript_str[start_pos+1:end_pos]
        if feature_type in ["transcript", "mRNA"]:
            if transcript_id in transcript_ids:
                logger.warning("Duplicated transcript id %s on line %d" % (transcript_id, line_count))
                gtf_correct = False
                transcript_ids[transcript_id] += 1
                transcript_id += ".%d" % transcript_ids[transcript_id]
            else:
                transcript_ids[transcript_id] = 0
        elif transcript_id in transcript_ids and transcript_ids[transcript_id] > 0:
            transcript_id += ".%d" % transcript_ids[transcript_id]

        if gene_id == transcript_id:
            logger.warning("Transcript id and gene id are identical (%s) at line %d"  % (transcript_id, line_count))
            gtf_correct = False
            transcript_id += ".RNA.IsoQuant_corrected"

        new_attrs = []
        for i in range(len(attrs)):
            if i == gene_id_pos + 1:
                new_attrs.append('"%s";' % gene_id)
            elif feature_type != "gene" and i == transcript_id_pos + 1:
                new_attrs.append('"%s";' % transcript_id)
            else:
                new_attrs.append(attrs[i])
        new_line = "\t".join(v[:8] + [" ".join(new_attrs)]) + "\n"
        corrected_gtf += new_line

        if feature_type == "exon":
            exon_gene_ids.add(gene_id)
            exon_transcript_ids.add(transcript_id)

    complete_genedb = 1
    if len(transcript_ids) == 0 or len(gene_ids) == 0:
        complete_genedb = -1
    elif len(transcript_ids) < len(exon_transcript_ids) or len(gene_ids) < len(exon_gene_ids):
        complete_genedb = 0

    return gtf_correct, corrected_gtf, gtf_name + ".corrected" + inner_ext.lower(), complete_genedb


def check_gff3_duplicates(handle):
    gtf_correct = True
    gene_count = 0
    transcript_count = 0
    line_count = 0
    feature_ids = {}

    for l in handle.readlines():
        line_count += 1
        if l.startswith("#"):
            continue
        v = l.strip().split("\t")
        if len(v) < 9:
            continue

        feature_type = v[2]
        if feature_type == 'gene':
            gene_count += 1
        elif feature_type in ["transcript", "mRNA"]:
            transcript_count += 1

        attrs = v[8].split(";")
        id_pos = -1
        for i in range(len(attrs)):
            if attrs[i].startswith('ID'):
                id_pos = i
        if id_pos == -1:
            if feature_type in ["gene", "transcript", "mRNA"]:
                logger.warning("Malformed GTF line %d (ID attribute value cannot be found)" % line_count)
                logger.warning(l.strip())
                gtf_correct = False
            continue

        id_str = attrs[id_pos]
        id_value = id_str.split("=")[1]
        if id_value in feature_ids:
            logger.warning("Duplicated ID %s on line %d" % (id_value, line_count))
            gtf_correct = False
            feature_ids[id_value] += 1

    complete_genedb = 1
    if transcript_count == 0 or gene_count == 0:
        complete_genedb = -1

    return gtf_correct, None, None, complete_genedb


def find_converted_db(converted_gtfs, gtf_filename, complete_genedb):
    gtf_mtime = converted_gtfs.get(gtf_filename, {}).get('gtf_mtime')
    db_mtime = converted_gtfs.get(gtf_filename, {}).get('db_mtime')
    db_file = converted_gtfs.get(gtf_filename, {}).get('genedb')
    is_complete = converted_gtfs.get(gtf_filename, {}).get('complete_db')
    if (os.path.exists(gtf_filename) and os.path.getmtime(gtf_filename) == gtf_mtime and
            os.path.exists(db_file) and os.path.getmtime(db_file) == db_mtime and complete_genedb == is_complete):
        return db_file
    return None


def compare_stored_gtf(converted_gtfs, gtf_filename, genedb_filename):
    gtf_mtime = converted_gtfs.get(gtf_filename, {}).get('gtf_mtime')
    db_mtime = converted_gtfs.get(gtf_filename, {}).get('db_mtime')
    return (os.path.exists(gtf_filename) and os.path.getmtime(gtf_filename) == gtf_mtime and
            os.path.exists(genedb_filename) and os.path.getmtime(genedb_filename) == db_mtime)


def convert_db(gtf_filename, genedb_filename, convert_fn, args):
    genedb_filename = os.path.abspath(genedb_filename)

    with open(args.db_config_path, 'r') as f_in:
        converted_gtfs = json.load(f_in)

    if not args.clean_start:
        if convert_fn == gtf2db:
            converted_db = find_converted_db(converted_gtfs, gtf_filename, args.complete_genedb)
            if converted_db is not None:
                logger.info("Gene annotation file found. Using " + converted_db)
                return gtf_filename, converted_db
        else:
            for converted_gtf in converted_gtfs:
                if compare_stored_gtf(converted_gtfs, converted_gtf, genedb_filename):
                    logger.info("Gene annotation file found. Using " + converted_gtf)
                    return converted_gtf, genedb_filename

    if convert_fn == gtf2db:
        convert_fn(gtf_filename, genedb_filename, args.complete_genedb, args.gtf_check)
    else:
        convert_fn(genedb_filename, gtf_filename)
    converted_gtfs[gtf_filename] = {
        'genedb': genedb_filename,
        'gtf_mtime': os.path.getmtime(gtf_filename),
        'db_mtime': os.path.getmtime(genedb_filename),
        'complete_db': args.complete_genedb
    }
    with open(args.db_config_path, 'w') as f_out:
        json.dump(converted_gtfs, f_out)
    return gtf_filename, genedb_filename


mode_dict = {"gtf2db" : gtf2db, "db2gtf" : db2gtf, "db2bed" : db2bed}


def parse_args():
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument("--output", "-o", type=str, help="output file", required=True)
    parser.add_argument("--input", "-i", type=str, help="input gtf/db file", required=True)
    parser.add_argument('--complete_genedb', '-c', action='store_true', default=False,
                        help="use this flag if gene annotation contains transcript and gene metafeatures, "
                             "e.g. with official annotations, such as GENCODE; "
                             "speeds up gene database conversion")
    parser.add_argument("--mode", "-m", type=str, default="gtf2db", help="[gtf2db|db2gtf|db2bed]")


    args = parser.parse_args()
    if not args.output or not args.input:
        parser.print_usage()
        sys.exit(IsoQuantExitCode.MISSING_REQUIRED_OPTION)
    return args


def main():
    args = parse_args()
    if args.mode in mode_dict:
        conversion_func = mode_dict[args.mode]
    else:
        logger.error("Mode %s is not available" % args.mode)
        sys.exit(IsoQuantExitCode.INVALID_CONVERSION_MODE)

    conversion_func(args.input, args.output, args.complete_genedb)


if __name__ == "__main__":
    try:
        main()
    except SystemExit:
        raise
    except:
        print_exc()




############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# All Rights Reserved
# See file LICENSE for details.
############################################################################

import pysam
import sys
import argparse
import gffutils
import time
import networkx as nx
from networkx.algorithms import bipartite
from enum import Enum, unique

from .common import *
from .gtf2db import *

ABSENT_PAIR = (-1,-1)

@unique
class IntronType(Enum):
	equal = 1
	overlap = 4
	only_short = 2
	only_isoquant = 3
	short_contains_iso = 6
	iso_contains_short = 5
	
@unique
class IntronReference(Enum):
	short_right = 1
	iso_right = 2
	both = 3
	none = 4

# just doing everything here now to avoid messing with anything
def parse_args():
	parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter)
	parser.add_argument("--short", "-s", help="input sam/bam file containing alignment of the short reads", 
						type=str, dest = "short", required = True)
	parser.add_argument("--isoquant", "-i", help="input isoquant annotation", 
						type=str, dest = "iso", required = True)
	parser.add_argument("--reference", "-r", help="input reference gtf file",
                        type=str, dest = "ref", default = "")
	parser.add_argument("--output", "-o", help="output folder location",
						type=str, dest = "out", required = True)
	return parser.parse_args()

def get_introns_old(f):
	samfile = pysam.AlignmentFile(f, "rb") 
	introns = samfile.find_introns(read for read in samfile.fetch())
	samfile.close()
	i_list = set()
	for i in introns.keys():
		#i_list.append((i[0]+1,i[1]+1))
		i_list.add(i)
	return i_list
	
def get_introns(f, chromosome, start, end):
		samfile = pysam.AlignmentFile(f, "rb") 
		intr = samfile.find_introns(samfile.fetch(self.chromosome, start = self.start, stop = self.end))
		samfile.close()
		i_list = set()
		for i in intr.keys():
			if(type(i)!="int"): 
				i_list.add((i[0]+1,i[1]))
		return i_list	

def introns_from_gene(db, gene, f):
	all_isoforms_introns = {}
	all_isoforms_exons = {}
	s = gene.start
	end = gene.end  # probably not just in the gene right? maybe also slightly before and slightly after?
	
	for t in db.children(gene, featuretype=('transcript', 'mRNA'), order_by='start'):
		all_isoforms_exons[t.id] = []
		for e in db.children(t, order_by='start'):
			if e.featuretype == 'exon':
				all_isoforms_exons[t.id].append((e.start, e.end))
				
		all_isoforms_introns[t.id] = junctions_from_blocks(all_isoforms_exons[t.id])
		
	introns = []
	for i in all_isoforms_introns.keys():
		introns.extend(all_isoforms_introns[i])
	
	samfile = pysam.AlignmentFile(f, "rb") 
	intr = samfile.find_introns(samfile.fetch(start = s, stop = end))
	samfile.close()
	i_list = []
	for i in intr.keys():
		#i_list.append((i[0]+1,i[1]+1))
		if(type(i)!="int"): 
			i_list.append((i[0]+1,i[1]))
		
	return introns, i_list
	
def introns_from_region(db, gene_list, current_region, f): # Kann ich das aus der reference auch selektiv rausbekommen? Vielleicht kann ich mir ne Liste ausgeben lassen innerhalb bestimmter Grenzen
	
	chromosome = gene_list[0].seqid
	
	all_isoforms_introns = {}
	all_isoforms_exons = {}
	
	for gene_db in gene_list:
		for t in db.children(gene_db, featuretype=('transcript', 'mRNA'), order_by='start'):
			all_isoforms_exons[t.id] = []
			for e in db.children(t, order_by='start'):
				if e.featuretype == 'exon':
					all_isoforms_exons[t.id].append((e.start, e.end))

			all_isoforms_introns[t.id] = junctions_from_blocks(all_isoforms_exons[t.id])
	
	introns = set()
	for i in all_isoforms_introns.keys():
		introns.update(all_isoforms_introns[i])
		
	samfile = pysam.AlignmentFile(f, "rb") 
	intr = samfile.find_introns(samfile.fetch(chromosome, start = current_region[0], stop = current_region[1]))
	samfile.close()
	i_list = set()
	for i in intr.keys():
		if(type(i)!="int"): 
			i_list.add((i[0]+1,i[1]))
	return introns, i_list, intr
	
def ref_from_region(reference, region): 
	
	gene_list = list(reference.features_of_type('gene', limit = region, order_by=('seqid', 'start')))
	
	all_isoforms_introns = {}
	all_isoforms_exons = {}
	
	for gene_db in gene_list:
		if gene_db.seqid == chromosome:
			for t in reference.children(gene_db, featuretype=('transcript', 'mRNA'), order_by='start'):
				all_isoforms_exons[t.id] = []
				for e in reference.children(t, order_by='start'):
					if e.featuretype == 'exon':
						all_isoforms_exons[t.id].append((e.start, e.end))

				all_isoforms_introns[t.id] = junctions_from_blocks(all_isoforms_exons[t.id])
	
	introns = set()
	for i in all_isoforms_introns.keys():
		introns.update(all_isoforms_introns[i])
		#print(i)
		#print(all_isoforms_introns[i])
		
	return introns

# from gene_info.py
def introns_from_db(db):
	gene_list = list(db.all_features(featuretype="gene"))
	#print(gene_list)
	all_isoforms_introns = {}
	all_isoforms_exons = {}
	for gene_db in gene_list:
		for t in db.children(gene_db, featuretype=('transcript', 'mRNA'), order_by='start'):
			all_isoforms_exons[t.id] = []
			for e in db.children(t, order_by='start'):
				if e.featuretype == 'exon':
					all_isoforms_exons[t.id].append((e.start, e.end))

			all_isoforms_introns[t.id] = junctions_from_blocks(all_isoforms_exons[t.id])

	if db and not all_isoforms_exons:
		logger.warning("Gene %s has no exons / transcripts, check your input annotation" % gene_list[0].id)
	
	introns = set()
	for i in all_isoforms_introns.keys():
		introns.update(all_isoforms_introns[i])
	
	return introns

def compare_introns(short, iso):
	classification = {}
	found_short = [False]*len(short)
	found_iso = [False]*len(iso)
	#print(len(short))
	#print(len(iso))
	for i in range(len(short)): 
		for j in range(len(iso)):
			#print(short[i][1])
			#print(iso[j][0])
			if short[i][1] < iso[j][0] or short[i][0] > iso[j][1]:
				#print("done")
				continue
			elif short[i] == iso[j]:
				classification[(short[i],iso[j])] = IntronType.equal
				found_short[i] = True
				found_iso[j] = True
				#print("equal")
				continue
			elif (short[i][0] <= iso[j][0] and short[i][1] <= iso[j][1]) or (short[i][0] >= iso[j][0] and short[i][1] >= iso[j][1]):
				classification[(short[i],iso[j])] = IntronType.overlap
				found_short[i] = True
				found_iso[j] = True
				#print("overlap")
				continue
			elif short[i][0] >= iso[j][0] and short[i][1] <= iso[j][1]:
				classification[(short[i],iso[j])] = IntronType.iso_contains_short
				found_short[i] = True
				found_iso[j] = True
				#print("short contained in iso")
				continue
			elif short[i][0] <= iso[j][0] and short[i][1] >= iso[j][1]:
				classification[(short[i],iso[j])] = IntronType.short_contains_iso
				found_short[i] = True
				found_iso[j] = True
				#print("iso contained in short")
				continue
		if not found_short[i]:
			classification[(short[i],ABSENT_PAIR)] = IntronType.only_short
			#print("not found in iso")
	for j in range(len(iso)):
		if not found_iso[j]:
			classification[(ABSENT_PAIR,iso[j])] = IntronType.only_isoquant
			#print("not found in short")
	return classification
	
def classify_introns(iso, short):
	if (short[0] <= iso[0] and short[1] <= iso[1]) or (short[0] >= iso[0] and short[1] >= iso[1]):
		return IntronType.overlap
	elif short[0] >= iso[0] and short[1] <= iso[1]:
		return IntronType.iso_contains_short
	elif short[0] <= iso[0] and short[1] >= iso[1]:
		return IntronType.short_contains_iso
	return IntronType.only_isoquant
	
def classify_reference(iso, short, ref, chromosome, counts):
	#if iso in ref and any(sh in ref for sh in short):
	if iso in ref and short in ref:
		print("both:", chromosome, ":" , iso, ",", short, "count:", counts[(short[0]-1,short[1])]) 
		return IntronReference.both
	elif iso in ref:
		print("iso:", chromosome, ":" , iso, ",", short, "count:", counts[(short[0]-1,short[1])])
		return IntronReference.iso_right
	#elif any(sh in ref for sh in short):
	elif short in ref:
		print("short:", chromosome, ":" , iso, ",", short, "count:", counts[(short[0]-1,short[1])])
		return IntronReference.short_right
	else:
		print("none:", chromosome, ":" , iso, ",", short, "count:", counts[(short[0]-1,short[1])])
		return IntronReference.none
	
def compare_alg(short, iso, counts, ref, chromosome):
	equal = iso.intersection(short)
	short = short.difference(equal)
	iso = iso.difference(equal)
	G = nx.Graph()
	G.add_nodes_from(iso, bipartite = 0)
	G.add_nodes_from(short, bipartite = 1)
	score = 1000000000000
	sh = (0,0)
	#score = [1000000000000,1000000000000,1000000000000]
	#sh = [(0,0),(0,0),(0,0)]
	chosen = []
	additional = []
	classification = {}
	cl_ref = {}
	scounts = []
	for i in iso:
		for s in short: 
			x = abs(i[0] - s[0]) + abs(i[1] - s[1])
			if (overlaps(i, s) and abs(i[0] - s[0]) <= 12 and abs(i[1] - s[1]) <= 12):
				# if x < score[0]:
					# score[0] = x
					# sh[0] = s
				# elif x < score[1]:
					# score[1] = x
					# sh[1] = s
				# elif x < score[2]:
					# score[2] = x
					# sh[2] = s
				if x < score:
					score = x
					sh = s
					additional = []
				if x == score and not s == sh:
					additional.append(s)
				G.add_edge(i, s, weight = abs(i[0] - s[0]) + abs(i[1] - s[1]))
		#classification[(i,sh[0])] = classify_introns(i,sh[0])
		#classification[(i,sh[1])] = classify_introns(i,sh[1])
		#classification[(i,sh[2])] = classify_introns(i,sh[2])
		#classification[(i,sh)] = classify_introns(i,sh)
		classify_reference(i, sh, ref, chromosome, counts)
		for x in additional:
			classify_reference(i, x, ref, chromosome, counts)
		# for y in sh:
			# if(not y == (0,0)):
				# chosen.append(y)
				# scounts.append(counts[(y[0]-1,y[1])])
		if(not sh == (0,0)):
			chosen.append(sh)
			#scounts.append(counts[(sh[0]-1,sh[1])])
		#score = [1000000000000,1000000000000,1000000000000]
		#sh = [(0,0),(0,0),(0,0)]
		score = 1000000000000
		sh = (0,0)
	eqcounts = []
	#for i in equal:
	#	classification[(i,i)] = IntronType.equal
	#	eqcounts.append(counts[(i[0]-1,i[1])])
	eq = len(equal)
	extra = short.difference(chosen)
	excounts = []
	#for i in extra:
		#excounts.append(counts[(i[0]-1,i[1])])
	
	ex_only = len(extra.intersection(ref))		
	#print("Counts in equal introns: ", eqcounts)
	#print("Counts in extra introns: ", excounts)
	#print("Counts in paired introns: ", scounts)
	return classification, eq, excounts, scounts, ex_only


start = time.time()
args = parse_args()
#print(args.short)
short_int = get_introns(args.short)
#print(short_int[1])
if(args.iso[-2:] != "db"):
	gtf = args.iso
	args.output_exists = os.path.exists(args.out)
	if not args.output_exists:
		os.makedirs(args.out)
	db_file = os.path.join(args.out, "mouse_isoquant.db")
	gtf2db(gtf, db_file, True)
else:
	db_file = args.iso
if(args.ref[-2:] != "db"):
	gtf = args.ref
	args.output_exists = os.path.exists(args.out)
	if not args.output_exists:
		os.makedirs(args.out)
	ref = os.path.join(args.out, "mouse_reference.db")
	gtf2db(gtf, ref, True)
else:
	ref = args.ref
db = gffutils.FeatureDB(db_file)
ref_db = gffutils.FeatureDB(ref)
reference = introns_from_db(ref_db)
gene_list = list(db.features_of_type('gene', order_by=('seqid', 'start')))
cl = {}
onlyshort = 0
current_region = (0,0)
chromosome = 0
genes = []
equal_c = 0
extra_c = []
paired_c = []
for gene in gene_list:
	if overlaps((gene.start,gene.end), current_region) and gene.seqid == chromosome: # Chromosom mit abfragen?
		genes.append(gene)
		current_region = (current_region[0], max(current_region[1], gene.end))
	else:
		if genes : 
			iso_int, short_int, counts = introns_from_region(db, genes, current_region, args.short)
			ref_int = ref_from_region(ref_db, (chromosome, current_region[0], current_region[1]))
			#print(len(iso_int.difference(ref_int)))
			c, eq, ex, p, s = compare_alg(short_int, iso_int, counts, ref_int, chromosome)
			cl.update(c)
			#equal_c.extend(eq)
			equal_c += eq
			extra_c.extend(ex)
			paired_c.extend(p)
			onlyshort += s
		genes = [gene]
		current_region = (gene.start, gene.end)
		chromosome = gene.seqid
# final dataset
print("Number of correct introns only found in Illumina:", onlyshort)
print("Number of Introns equal in IsoQuant and Illumina:", equal_c)
end = time.time()
print(end - start)




############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# Copyright (c) 2020-2022 Saint Petersburg State University
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import logging
import os
import re
import subprocess
import sys
import math
from collections import defaultdict
from enum import Enum

logger = logging.getLogger('IsoQuant')


def _get_log_params():
    """Return (log_file, log_level) from the current IsoQuant logger configuration."""
    w = logging.getLogger('IsoQuant')
    log_level = w.level or logging.INFO
    log_file = None
    for h in w.handlers:
        if hasattr(h, 'baseFilename'):
            log_file = h.baseFilename
            break
    return log_file, log_level


def setup_worker_logging(log_file, log_level):
    """Initialize IsoQuant logging in a worker process.

    Must be passed as the 'initializer' argument to every ProcessPoolExecutor.
    Works for both 'fork' (clears broken inherited handlers) and 'spawn'
    (creates handlers from scratch). Python 3.14 changed the default start
    method from 'fork' to 'spawn' on Linux, and also changed _at_fork_reinit
    behaviour so inherited handlers can break even with explicit fork.
    """
    w = logging.getLogger('IsoQuant')
    for h in w.handlers[:]:
        w.removeHandler(h)
    w.setLevel(log_level)
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    if log_file:
        fh = logging.FileHandler(log_file, mode='a')
        fh.set_name("isoquant_file_log")
        fh.setLevel(log_level)
        fh.setFormatter(formatter)
        w.addHandler(fh)
    ch = logging.StreamHandler(sys.stdout)
    ch.set_name("isoquant_screen_log")
    ch.setLevel(logging.INFO)
    ch.setFormatter(formatter)
    w.addHandler(ch)


class CigarEvent(Enum):
    match = 0
    insertion = 1
    deletion = 2
    skipped = 3
    soft_clipping = 4
    hard_clipping = 5
    padding = 6
    seq_match = 7
    seq_mismatch = 8

    @classmethod
    def get_match_events(cls):
        return {cls.match, cls.seq_match, cls.seq_mismatch}

    @classmethod
    def get_ins_del_match_events(cls):
        return {cls.match, cls.insertion, cls.deletion, cls.seq_match, cls.seq_mismatch}


class TranscriptNaming:
    transcript_prefix = "transcript"
    novel_gene_prefix = "novel_gene_"
    nic_transcript_suffix = ".nic"
    nnic_transcript_suffix = ".nnic"


# key, value
def get_first_best_from_sorted(sorted_list_of_pairs):
    if not sorted_list_of_pairs:
        return []
    best_value = sorted_list_of_pairs[0][1]
    result = []
    for x in sorted_list_of_pairs:
        if x[1] > best_value:
            break
        result.append(x[0])
    return result


def list_to_str(element_list, element_delim=','):
    if len(element_list) == 0:
        return "."
    return element_delim.join(list(map(str, element_list)))


def rindex(l, el):
    for i in range(len(l) - 1, -1, -1):
        if l[i] == el:
            return i
    raise ValueError(str(el) + " is not in list")


def get_top_count(dict):
    if not dict:
        return None
    return max(dict.items(), key=lambda x: x[1])[0]


def find_closest(value, value_list):
    if not value_list:
        return None, math.inf
    best_diff = math.inf
    best_el = None
    for v in value_list:
        diff = abs(v - value)
        if diff < best_diff:
            best_diff = diff
            best_el = v
    return best_el, best_diff


def rreplace(s, old, new):
    return new.join(s.rsplit(old, 1))


# dict -> counts dict
def get_best_from_count_dicts(dict):
    res = {}
    for k in dict.keys():
        res[k] = get_top_count(dict[k])
    return res


def get_collective_property(value_collection, property_dict):
    counts = defaultdict(int)
    for v in value_collection:
        if v not in property_dict:
            continue
        counts[property_dict[v]] += 1
    return get_top_count(counts)


def proper_plural_form(name, count):
    return str(count) + " " + name + ("" if count == 1 else "s")


# check whether genes overlap and should be processed together
def genes_overlap(gene_db1, gene_db2):
    if gene_db1.seqid != gene_db2.seqid:
        return False
    return overlaps((gene_db1.start, gene_db1.end), (gene_db2.start, gene_db2.end))


def genes_contain(gene_db1, gene_db2):
    if gene_db1.seqid != gene_db2.seqid:
        return False
    return contains((gene_db1.start, gene_db1.end), (gene_db2.start, gene_db2.end))


def argmin(l):
    if len(l) == 0:
        return -1
    min_v = l[0]
    min_i = 0
    for i in range(len(l)):
        if l[i] < min_v:
            min_v = l[i]
            min_i = i
    return min_i


def cmp(x,y):
    if x<y:
        return -1
    elif x>y:
        return 1
    return 0


# == range operations ==
def overlaps(range1, range2):
    return not (range1[1] < range2[0] or range1[0] > range2[1])


def overlap_intervals(range1, range2):
    return max(range1[0], range2[0]), min(range1[1], range2[1])


def overlaps_at_least(range1, range2, delta=0):
    ovlp1 = range1[1] - range2[0]
    ovlp2 = range2[1] - range1[0]
    if ovlp1 < 0 or ovlp2 < 0:
        return False
    d = delta - 1
    if range1[1] < range2[1]:
        return ovlp1 >= d or range1[0] >= range2[0]
    else:
        return ovlp2 >= d or range1[0] <= range2[0]


# dangerous function, works only when range1 and range2 are already known to overlap, do not use if unsure
def overlaps_at_least_when_overlap(range1, range2, delta=0):
    if range1[1] < range2[1]:
        return range1[0] >= range2[0] or range1[1] - range2[0] + 1 >= delta
    else:
        return range1[0] <= range2[0] or range2[1] - range1[0] + 1 >= delta


def intersection_len(range1, range2):
    return max(0, min(range1[1], range2[1]) - max(range1[0], range2[0]) + 1)


def left_of(range1, range2):
    return range1[1] < range2[0]


def equal_ranges(range1, range2, delta=0):
    return abs(range1[0] - range2[0]) <= delta and abs(range1[1] - range2[1]) <= delta


def covers_end(bigger_range, smaller_range):
    return bigger_range[0] <= smaller_range[0] <= bigger_range[1] <= smaller_range[1]


def covers_start(bigger_range, smaller_range):
    return smaller_range[0] <= bigger_range[0] <= smaller_range[1] <= bigger_range[1]


def contains(bigger_range, smaller_range):
    return bigger_range[1] >= smaller_range[1] and bigger_range[0] <= smaller_range[0]


def contains_well_inside(bigger_range, smaller_range, delta=1):
    return bigger_range[1] >= smaller_range[1]+delta and bigger_range[0] <= smaller_range[0]-delta


def contains_approx(bigger_range, smaller_range, delta = 1):
    return bigger_range[1] + delta >= smaller_range[1] and bigger_range[0] - delta <= smaller_range[0]


def range_list_to_str(range_list, element_delim=',', coord_delim='-'):
    return element_delim.join(list(map(lambda x: str(x[0]) + coord_delim + str(x[1]), range_list)))


def max_range(range1, range2):
    return (min(range1[0], range2[0]), max(range1[1], range2[1]))


def interval_len(interval):
    return interval[1] - interval[0] + 1


def intervals_total_length(sorted_range_list):
    total_len = 0
    for r in sorted_range_list:
        total_len += interval_len(r)
    return total_len


# check that all exons are sorted and have correct coordinates
def validate_exons(self, exons):
    return exons == sorted(exons) and all(0 < x[0] <= x[1] for x in exons)


# sum up all inervals until pos, pos not included
def sum_intervals_to_point(sorted_range_list, pos):
    if pos <= sorted_range_list[0][0]:
        return 0
    elif pos > sorted_range_list[-1][1]:
        return intervals_total_length(sorted_range_list)

    i = 0
    total_len = 0
    while i < len(sorted_range_list) and sorted_range_list[i][0] < pos:
        if sorted_range_list[i][0] <= pos <= sorted_range_list[i][1]:
            total_len += pos - sorted_range_list[i][0]
        else:
            total_len += sorted_range_list[i][1] - sorted_range_list[i][0] + 1
        i += 1
    return total_len


# sum up all intervals from pos, pos not included
def sum_intervals_from_point(sorted_range_list, pos):
    if pos < sorted_range_list[0][0]:
        return intervals_total_length(sorted_range_list)
    elif pos > sorted_range_list[-1][1]:
        return 0

    i = len(sorted_range_list) - 1
    total_len = 0
    while i >= 0 and sorted_range_list[i][1] > pos:
        if sorted_range_list[i][0] <= pos <= sorted_range_list[i][1]:
            total_len += sorted_range_list[i][1] - pos
        else:
            total_len += sorted_range_list[i][1] - sorted_range_list[i][0] + 1
        i -= 1
    return total_len


def jaccard_similarity(sorted_range_list1, sorted_range_list2):
    union = 0
    intersection = 0
    pos1 = 0
    pos2 = 0
    included1 = [0 for i in range(len(sorted_range_list1))]
    included2 = [0 for i in range(len(sorted_range_list2))]

    while pos1 < len(sorted_range_list1) and pos2 < len(sorted_range_list2):
        block1 = sorted_range_list1[pos1]
        block2 = sorted_range_list2[pos2]
        if overlaps(block1, block2):
            assert (included2[pos2] == 0 or included1[pos1] == 0)

            intersection += min(block1[1], block2[1]) - max(block1[0], block2[0]) + 1
            if included2[pos2] == 0 and included1[pos1] == 0:
                # both blocks were not counted
                union += max(block1[1], block2[1]) - min(block1[0], block2[0]) + 1
            elif included2[pos2] == 1:
                # second block was included, take extra bite from block1
                union += max(0, block1[1] - block2[1])
            else:
                # first block was included, take extra bite from block2
                union += max(0, block2[1] - block1[1])

            included1[pos1] = 1
            included2[pos2] = 1
            if block2[1] < block1[1]:
                pos2 += 1
            else:
                pos1 += 1
        elif left_of(block2, block1):
            if included2[pos2] == 0:
                union += block2[1] - block2[0] + 1
                included2[pos2] = 1
            pos2 += 1
        else:
            if included1[pos1] == 0:
                union += block1[1] - block1[0] + 1
                included1[pos1] = 1
            pos1 += 1

    while pos1 < len(sorted_range_list1):
        if included1[pos1] == 0:
            union += sorted_range_list1[pos1][1] - sorted_range_list1[pos1][0] + 1
        pos1 += 1

    while pos2 < len(sorted_range_list2):
        if included2[pos2] == 0:
            union += sorted_range_list2[pos2][1] - sorted_range_list2[pos2][0] + 1
        pos2 += 1

    assert (union != 0)
    return float(intersection) / float(union)


def merge_ranges(sorted_range_list1, sorted_range_list2):
    union = []
    intersection = 0
    pos1 = 0
    pos2 = 0
    included1 = [0 for i in range(len(sorted_range_list1))]
    included2 = [0 for i in range(len(sorted_range_list2))]

    while pos1 < len(sorted_range_list1) and pos2 < len(sorted_range_list2):
        block1 = sorted_range_list1[pos1]
        block2 = sorted_range_list2[pos2]
        if overlaps(block1, block2):
            assert (included2[pos2] == 0 or included1[pos1] == 0)

            intersection += min(block1[1], block2[1]) - max(block1[0], block2[0]) + 1
            if included2[pos2] == 0 and included1[pos1] == 0:
                # both blocks were not counted
                union.append((min(block1[0], block2[0]), max(block1[1], block2[1])))
            elif included2[pos2] == 1:
                # second block was included, take extra bite from block1
                last_block = union[-1]
                union[-1] = (last_block[0], max(last_block[1], block1[1]))
            else:
                # first block was included, take extra bite from block2
                last_block = union[-1]
                union[-1] = (last_block[0], max(last_block[1], block2[1]))

            included1[pos1] = 1
            included2[pos2] = 1
            if block2[1] < block1[1]:
                pos2 += 1
            else:
                pos1 += 1
        elif left_of(block2, block1):
            if included2[pos2] == 0:
                union.append(block2)
                included2[pos2] = 1
            pos2 += 1
        else:
            if included1[pos1] == 0:
                union.append(block1)
                included1[pos1] = 1
            pos1 += 1

    assert pos1 == len(sorted_range_list1) or pos2 == len(sorted_range_list2)
    while pos1 < len(sorted_range_list1):
        if included1[pos1] == 0:
            union.append(sorted_range_list1[pos1])
        pos1 += 1
    while pos2 < len(sorted_range_list2):
        if included2[pos2] == 0:
            union.append(sorted_range_list2[pos2])
        pos2 += 1

    assert len(union) != 0
    return union


def read_coverage_fraction(read_range_list, isoform_range_list):
    intersection = 0
    pos1 = 0
    pos2 = 0

    while pos1 < len(read_range_list) and pos2 < len(isoform_range_list):
        block1 = read_range_list[pos1]
        block2 = isoform_range_list[pos2]
        if overlaps(block1, block2):
            intersection += min(block1[1], block2[1]) - max(block1[0], block2[0]) + 1
            if block2[1] < block1[1]:
                pos2 += 1
            else:
                pos1 += 1
        elif left_of(block2, block1):
            pos2 += 1
        else:
            pos1 += 1

    read_length = intervals_total_length(read_range_list)
    return float(intersection) / float(read_length)


def extra_exon_percentage(isoform_region, read_exons):
    total_read_len = 0
    outside_read_len = 0
    for e in read_exons:
        total_read_len += e[1] - e[0] + 1
        if e[0] < isoform_region[0]:
            outside_read_len += min(e[1], isoform_region[0] - 1) - e[0] + 1
        if e[1] > isoform_region[1]:
            outside_read_len += e[1] - max(e[0], isoform_region[1] + 1) + 1

    return float(outside_read_len) / float(total_read_len)


# == working with alignment blocks ==
def junctions_from_blocks(sorted_blocks):
    junctions = []
    if len(sorted_blocks) >= 2:
        for i in range(0, len(sorted_blocks) - 1):
            if sorted_blocks[i][1] + 1 < sorted_blocks[i + 1][0]:
                junctions.append((sorted_blocks[i][1] + 1, sorted_blocks[i + 1][0] - 1))
    return junctions


def get_following_exon_from_junctions(region, introns, intron_position):
    if intron_position == len(introns) - 1 or intron_position == -1:
        # intron precedes the last exon
        following_exon_end = region[1]
    else:
        following_exon_end = introns[intron_position + 1][0] - 1
    return (introns[intron_position][1] + 1, following_exon_end)


def get_preceding_exon_from_junctions(region, introns, intron_position):
    assert intron_position <= len(introns)
    if intron_position == 0:
        # intron precedes the first exon
        preceding_exon_start = region[0]
    else:
        preceding_exon_start = introns[intron_position - 1][1] + 1

    if intron_position == len(introns):
        return (preceding_exon_start, region[1])
    return (preceding_exon_start, introns[intron_position][0] - 1)


def get_exon(read_region, read_junctions, exon_position):
    assert exon_position <= len(read_junctions)
    if exon_position < 0:
        exon_position = len(read_junctions) + exon_position + 1

    if exon_position == 0:
        return (read_region[0], read_junctions[0][0] - 1)
    elif exon_position == len(read_junctions):
        return (read_junctions[-1][1] + 1, read_region[1])
    else:
        return (read_junctions[exon_position - 1][1] + 1, read_junctions[exon_position][0] - 1)


def get_exons(read_region, read_introns):
    return junctions_from_blocks([(-math.inf, read_region[0] - 1)] + read_introns + [(read_region[1] + 1, math.inf)])


def concat_gapless_blocks(blocks, cigar_tuples):
    cigar_index = 0
    block_index = 0
    resulting_blocks = []

    current_block = None
    deletions_before_block = 0

    while cigar_index < len(cigar_tuples) and block_index < len(blocks):
        # init new block
        cigar_event = CigarEvent(cigar_tuples[cigar_index][0])
        if current_block is None:
            # init new block from match
            if cigar_event in CigarEvent.get_match_events():
                current_block = (blocks[block_index][0] - deletions_before_block, blocks[block_index][1])
                deletions_before_block = 0
                block_index += 1
            # keep track of deletions before matched block
            elif cigar_event == CigarEvent.deletion:
                deletions_before_block = cigar_tuples[cigar_index][1]
        # found intron, add current block
        elif cigar_event == CigarEvent.skipped:
            resulting_blocks.append(current_block)
            current_block = None
        # add deletion to block
        elif cigar_event == CigarEvent.deletion:
            current_block = (current_block[0], current_block[1] + cigar_tuples[cigar_index][1])
        # found match - merge blocks
        elif cigar_event in CigarEvent.get_match_events():
            # if abs(current_block[1] - blocks[block_index][0]) > 1:
            #    logger.debug("Distant blocks")
            #    logger.debug(current_block, blocks[block_index])
            current_block = (current_block[0], blocks[block_index][1])

            block_index += 1
        cigar_index += 1

    if current_block is not None:
        resulting_blocks.append(current_block)

    return resulting_blocks


def get_read_blocks(ref_start, cigar_tuples):
    read_pos = 0
    ref_pos = ref_start + 1
    cigar_index = 0
    current_ref_block_start = None
    current_read_block_start = None
    current_cigar_block_start = None
    has_match = False
    ref_blocks = []
    cigar_blocks = []
    read_blocks = []

    while cigar_index < len(cigar_tuples):
        cigar_event = CigarEvent(cigar_tuples[cigar_index][0])
        event_len = cigar_tuples[cigar_index][1]

        if current_ref_block_start is None and cigar_event in CigarEvent.get_ins_del_match_events():
            # init new block from match
            current_ref_block_start = ref_pos
            current_read_block_start = read_pos
            current_cigar_block_start = cigar_index
            if cigar_event == CigarEvent.insertion:
                read_pos += event_len
            elif cigar_event == CigarEvent.deletion:
                ref_pos += event_len
            else:
                read_pos += event_len
                ref_pos += event_len
                has_match = True
        # found intron, add current block
        elif cigar_event in CigarEvent.get_match_events():
            read_pos += event_len
            ref_pos += event_len
            has_match = True
        elif cigar_event == CigarEvent.insertion:
            read_pos += event_len
        elif cigar_event == CigarEvent.deletion:
            ref_pos += event_len
        elif cigar_event == CigarEvent.skipped:
            if current_ref_block_start:
                if has_match:
                    ref_blocks.append((current_ref_block_start, ref_pos - 1))
                    read_blocks.append((current_read_block_start, read_pos - 1))
                    cigar_blocks.append((current_cigar_block_start, cigar_index - 1))
                has_match = False
                current_ref_block_start = None
                current_read_block_start = None
                current_cigar_block_start = None
            ref_pos += event_len
        elif cigar_event == CigarEvent.soft_clipping:
            if current_ref_block_start:
                if has_match:
                    ref_blocks.append((current_ref_block_start, ref_pos - 1))
                    read_blocks.append((current_read_block_start, read_pos - 1))
                    cigar_blocks.append((current_cigar_block_start, cigar_index - 1))
                has_match = False
                current_ref_block_start = None
                current_read_block_start = None
                current_cigar_block_start = None
            read_pos += event_len

        cigar_index += 1

    if current_ref_block_start and has_match:
        ref_blocks.append((current_ref_block_start, ref_pos - 1))
        read_blocks.append((current_read_block_start, read_pos - 1))
        cigar_blocks.append((current_cigar_block_start, cigar_index - 1))

    return ref_blocks, read_blocks, cigar_blocks


def correct_bam_coords(blocks):
    return list(map(lambda x: (x[0] + 1, x[1]), blocks))


def truncate_read_to_polya(read_exons, polya_pos, polyt_pos):
    end_index = len(read_exons) - 1
    end_position = read_exons[-1][1]
    if polya_pos != -1:
        end_position = polya_pos
        while end_index >= 0:
            if read_exons[end_index][0] < polya_pos:
                break
            end_index -= 1

    start_index = 0
    start_position = read_exons[0][0]
    if polyt_pos != -1:
        start_position = polyt_pos
        while start_index <= end_index:
            if read_exons[start_index][1] > polyt_pos:
                break
            start_index += 1

    if start_position == read_exons[0][0] and end_position == read_exons[-1][1]:
        return read_exons

    if start_index == end_index:
        return [(start_position, end_position)]

    new_exons = [(start_position, read_exons[start_index][1])]
    new_exons += read_exons[start_index+1:end_index]
    new_exons.append((read_exons[end_index][0], end_position))
    return new_exons


# == profile functions ==
def is_subprofile(short_isoform_profile, long_isoform_profile):
    assert len(short_isoform_profile) == len(long_isoform_profile)
    assert 0 not in short_isoform_profile

    short_range_start = None
    short_range_end = None
    for i in range(len(short_isoform_profile)):
        if short_isoform_profile[i] in {-1, 1}:
            if short_range_start is None:
                short_range_start = i
            short_range_end = i

    assert short_range_end is not None

    for i in range(short_range_start, short_range_end + 1):
        if short_isoform_profile[i] != long_isoform_profile[i]:
            return False
    return True


def difference_in_present_features(profile1, profile2, diff_limit=-1, profile_range=None):
    """ computes Hamming distance for two profiles

    Parameters
    ----------
    profile1: list(int)
    profile2: list(int)
    diff_limit: int
    profile_range: tuple

    Returns
    -------
    result: int
        number of different items in lists

    """
    assert len(profile1) == len(profile2)
    d = 0
    if diff_limit == -1:
        diff_limit = len(profile1) + 1
    if profile_range is None:
        profile_range = (0, len(profile1))

    for i in range(*profile_range):
        if profile2[i] == 0 or profile1[i] == 0:
            continue
        if profile1[i] != profile2[i]:
            d += 1
        if d > diff_limit:
            break
    return d


def equal_profiles_in_range(isoforom_profile, read_profile, profile_range):
    # return true if there is difference between 2 profiles
    # isoforom_profile must not contain zeroes
    for i in range(*profile_range):
        if read_profile[i] == 0:
            continue
        if isoforom_profile[i] != read_profile[i]:
            return False
    return True

def count_both_present_features(profile1, profile2):
    assert len(profile1) == len(profile2)
    d = 0
    for i in range(len(profile1)):
        if profile1[i] == profile2[i] == 1:
            d += 1
    return d


def all_features_present(isoform_profile, read_profile):
    assert len(isoform_profile) == len(read_profile)
    for i in range(len(isoform_profile)):
        if isoform_profile[i] == 1 and read_profile[i] != 1:
            return False
    return True


def find_matching_positions(profile1, profile2):
    assert len(profile1) == len(profile2)
    matches = [0 for i in range(len(profile1))]
    for i in range(len(profile1)):
        if profile1[i] == profile2[i]:
            matches[i] = 1
    return matches


def has_overlapping_features(profile1, profile2, profile_range=None):
    assert len(profile1) == len(profile2)
    if profile_range is None:
        profile_range = (0, len(profile1))

    for i in range(*profile_range):
        if profile1[i] == profile2[i] == 1:
            return True
    return False


def has_inconsistent_features(read_profile, gene_profile):
    assert len(read_profile) == len(gene_profile)
    for i in range(len(read_profile)):
        if read_profile[i] != gene_profile[i] and read_profile[i] != 0:
            return True
    return False


def mask_profile(read_profile, true_profile):
    assert len(read_profile) == len(true_profile)

    masked_profile = []
    for i in range(len(true_profile)):
        if true_profile[i] == 1:
            masked_profile.append(read_profile[i])
        else:
            masked_profile.append(0)
    return masked_profile


def get_blocks_from_profile(features, profile):
    assert len(features) == len(profile)
    profile_features = []
    for i in range(len(profile)):
        if profile[i] == 1:
            profile_features.append(features[i])
    return profile_features


# assumes there are no contradictions
def left_truncated(read_profile, isoform_profile):
    if 1 not in read_profile or 1 not in isoform_profile:
        return True
    return read_profile.index(1) > isoform_profile.index(1)


# assumes there are no contradictions
def right_truncated(read_profile, isoform_profile):
    if 1 not in read_profile or 1 not in isoform_profile:
        return True
    return rindex(read_profile, 1) < rindex(isoform_profile, 1)


def get_path_to_program(program, dirpath=None, min_version=None):
    """
    returns the path to an executable or None if it can't be found
    """
    def is_exe(fpath):
        if os.path.isfile(fpath) and os.access(fpath, os.X_OK):
            if not min_version or check_version(fpath, min_version):
                return True

    def check_version(fpath, min_version):
        p = subprocess.run([fpath, '--version'], capture_output=True)
        version_pattern = re.compile(r'(?P<major_version>\d+)\.(?P<minor_version>\d+)')
        v = version_pattern.search(str(p.stdout))
        if not v.group('major_version') or not v.group('minor_version'):
            return False
        version, minor_version = map(int, min_version.split('.'))
        if int(v.group('major_version')) == version and int(v.group('minor_version')) >= minor_version:
            return True

    if dirpath:
        exe_file = os.path.join(dirpath, program)
        if is_exe(exe_file):
            return exe_file

    for path in os.environ["PATH"].split(os.pathsep):
        exe_file = os.path.join(path, program)
        if is_exe(exe_file):
            return exe_file
    return None


def count_noncanonincal(introns, reference_region, strand, ref_region_start=0):
    count = 0
    for intron in introns:
        intron_left_pos = intron[0] - ref_region_start
        intron_right_pos = intron[1] - ref_region_start
        left_site = reference_region[intron_left_pos:intron_left_pos + 2]
        right_site = reference_region[intron_right_pos - 1:intron_right_pos + 1]
        if strand == '+':
            count += 0 if (left_site, right_site) == ("GT", "AG") else 1
        else:
            count += 0 if (left_site, right_site) == ("CT", "AC") else 1
    return count


# left site always first
CANONICAL_FWD_SITES = {("GT", "AG"), ("GC", "AG"), ("AT", "AC")}
CANONICAL_REV_SITES = {("CT", "AC"), ("CT", "GC"), ("GT", "AT")}


def get_intron_strand(intron, reference_region, ref_region_start=1):
    intron_left_pos = intron[0] - ref_region_start
    intron_right_pos = intron[1] - ref_region_start
    left_site = str(reference_region[intron_left_pos:intron_left_pos + 2]).upper()
    right_site = str(reference_region[intron_right_pos - 1:intron_right_pos + 1]).upper()
    is_fwd = (left_site, right_site) in CANONICAL_FWD_SITES
    is_rev = (left_site, right_site) in CANONICAL_REV_SITES

    if is_fwd == is_rev:
        return '.'
    return '+' if is_fwd else '-'


def get_strand(introns, reference_region, ref_region_start=1):
    if len(introns) == 0:
        return '.'

    count_fwd = 0
    count_rev = 0
    for intron in introns:
        intron_left_pos = intron[0] - ref_region_start
        intron_right_pos = intron[1] - ref_region_start
        left_site = str(reference_region[intron_left_pos:intron_left_pos + 2])
        right_site = str(reference_region[intron_right_pos - 1:intron_right_pos + 1])
        count_fwd += 1 if (left_site, right_site) in CANONICAL_FWD_SITES else 0
        count_rev += 1 if (left_site, right_site) in CANONICAL_REV_SITES else 0

    if count_fwd == count_rev:
        return '.'
    return '+' if count_rev < count_fwd else '-'


# binary search of a coordinate in ordered non-overlapping intervals
def interval_bin_search(ordered_intervals, pos):
    if pos > ordered_intervals[-1][1] or pos < ordered_intervals[0][0]:
        return -1

    s = len(ordered_intervals) - 1
    if pos >= ordered_intervals[-1][0]:
        return s

    ind = s // 2
    current_step = s // 2
    while not (ordered_intervals[ind][0] <= pos < ordered_intervals[ind + 1][0]):
        current_step = max(1, current_step // 2)
        if pos < ordered_intervals[ind][0]:
            ind -= current_step
        else:
            ind += current_step
    return ind


def interval_bin_search_rev(ordered_intervals, pos):
    if pos > ordered_intervals[-1][1] or pos < ordered_intervals[0][0]:
        return -1

    if pos <= ordered_intervals[0][1]:
        return 0

    s = len(ordered_intervals) - 1
    ind = s // 2
    current_step = s // 2
    while not (ordered_intervals[ind - 1][1] < pos <= ordered_intervals[ind][1]):
        current_step = max(1, current_step // 2)
        if pos > ordered_intervals[ind][1]:
            ind += current_step
        else:
            ind -= current_step
    return ind


base_comp = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A', 'N': 'N', 'a': 't', 'c': 'g', 'g': 'c', 't': 'a', 'n': 'n'}


def reverse_complement(seq):
    lms = list(map(lambda x: base_comp[x], seq))[::-1]
    return ''.join(lms)


def large_output_enabled(args, output_type):
    """Check if a large output type is enabled.

    Args:
        args: Parsed command-line arguments
        output_type: One of 'read_assignments', 'corrected_bed', 'read2transcripts', 'allinfo'

    Returns:
        True if the output type is enabled, False otherwise
    """
    if not hasattr(args, 'large_output') or not args.large_output:
        return False
    if "none" in args.large_output:
        return False
    return output_type in args.large_output


############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# Copyright (c) 2020-2022 Saint Petersburg State University
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import logging
from enum import Enum, unique
from functools import partial
from collections import OrderedDict, defaultdict

from .serialization import *
from .common import (
    contains,
    equal_ranges,
    get_intron_strand,
    intervals_total_length,
    overlaps,
    junctions_from_blocks
)
from .id_policy import SimpleIDDistributor

logger = logging.getLogger('IsoQuant')


@unique
class TranscriptModelType(Enum):
    known = 1
    novel_in_catalog = 2
    novel_not_in_catalog = 10


# simple class for storing all information needed for GFF
class TranscriptModel:
    def __init__(self, chr_id, strand, transcript_id, gene_id, exon_blocks, transcript_type,
                 other_features=None, source="IsoQuant"):
        self.chr_id = chr_id
        self.strand = strand
        self.transcript_id = transcript_id
        self.gene_id = gene_id
        self.exon_blocks = exon_blocks
        self.transcript_type = transcript_type
        self.source = source
        self.other_features = [] if not other_features else other_features
        self.additional_info = OrderedDict()
        self.intron_path = ()

    @classmethod
    def from_reference_transcript(cls, gene_info, isoform_id):
        transcript_model = cls.__new__(cls)
        transcript_model.chr_id = gene_info.chr_id
        transcript_model.strand = gene_info.isoform_strands[isoform_id]
        transcript_model.transcript_id = isoform_id
        transcript_model.gene_id = gene_info.gene_id_map[isoform_id]
        transcript_model.exon_blocks = gene_info.all_isoforms_exons[isoform_id]
        transcript_model.transcript_type = TranscriptModelType.known
        transcript_model.source = gene_info.sources[isoform_id]
        transcript_model.other_features = gene_info.other_features[isoform_id]
        transcript_model.additional_info = OrderedDict()
        transcript_model.intron_path = ()
        return transcript_model

    def get_start(self):
        return self.exon_blocks[0][0]

    def get_end(self):
        return self.exon_blocks[-1][1]

    def check_additional(self, attribute):
        return attribute in self.additional_info

    def add_additional_attribute(self, attribute, value):
        self.additional_info[attribute] = value

    def additional_attributes_str(self):
        return " ".join(['%s "%s";' % (k, v) for k,v in self.additional_info.items()])


# storage for feature profiles of all known isoforms of a gene or a set of overlapping genes
class FeatureProfiles:
    def __init__(self):
        self.profiles = {}
        self.profile_ranges = {}
        self.features = []

    def set_features(self, features):
        self.features = features

    def set_profiles(self, transcript_id, transcript_features, transcript_region, comaprator):
        self.profiles[transcript_id] = [-1] * len(self.features)
        current_profile = self.profiles[transcript_id]

        for pos, feature in enumerate(self.features):
            if not overlaps(feature, transcript_region):
                current_profile[pos] = -2

        pos = 0
        for feature in transcript_features:
            while pos < len(self.features) and not comaprator(feature, self.features[pos]):
                pos += 1
            while pos < len(self.features) and comaprator(feature, self.features[pos]):
                current_profile[pos] = 1
                pos += 1

        start_pos = 0
        while start_pos < len(current_profile) and current_profile[start_pos] < 1:
            start_pos += 1
        end_pos = len(current_profile) - 1
        while end_pos >= 0 and current_profile[end_pos] < 1:
            end_pos -= 1
        self.profile_ranges[transcript_id] = (start_pos, end_pos + 1)

    def print_debug(self):
        logger.debug(str(self.features))
        for t in self.profiles:
            logger.debug("Profile for isoform " + t)
            logger.debug(self.profiles[t])


# exon/intron info
class FeatureInfo:
    feature_id_counter = SimpleIDDistributor()
    def __init__(self, chr_id, start, end, strand, type, gene_ids):
        self.id = FeatureInfo.feature_id_counter.increment()
        self.chr_id = chr_id
        self.start = start
        self.end = end
        self.strand = strand
        self.type = type
        self.gene_ids = gene_ids
        #self.id = "%s_%d_%d_%s" % (self.chr_id, self.start, self.end, self.strand)

    @staticmethod
    def header():
        return "chr\tstart\tend\tstrand\tflags\tgene_ids"

    def to_str(self):
        return "%s\t%d\t%d\t%s\t%s\t%s" % (self.chr_id, self.start, self.end, self.strand, self.type, ",".join(self.gene_ids))


class GeneList:
    def __init__(self, gene_id_list, delta, chr_id, start, end):
        self.gene_id_set = set(gene_id_list)
        self.delta = delta
        self.chr_id = chr_id
        self.start = start
        self.end = end

    def merge(self, other):
        assert self.chr_id == other.chr_id
        self.gene_id_set.update(other.gene_id_set)
        self.start = min(self.start, other.start)
        self.end = max(self.end, other.end)

    def overlaps(self, other):
        return self.chr_id == other.chr_id and len(self.gene_id_set.intersection(other.gene_id_set)) > 0

    @classmethod
    def deserialize(cls, infile):
        gene_info = cls.__new__(cls)
        gene_info.delta = read_int(infile)
        gene_info.gene_id_set = set()

        gene_count = read_int(infile)
        for i in range(gene_count):
            gene_id = read_string(infile)
            gene_info.gene_id_set.add(gene_id)
        gene_info.chr_id = read_string(infile)
        gene_info.start = read_int(infile)
        gene_info.end = read_int(infile)

        return gene_info

    def serialize(self, outfile):
        write_int(self.delta, outfile)
        write_int(len(self.gene_id_set), outfile)
        for g in self.gene_id_set:
            write_string(g.id, outfile)
        write_string(self.chr_id, outfile)
        write_int(self.start, outfile)
        write_int(self.end, outfile)


# All gene(s) information
class GeneInfo:
    EXTRA_BASES_FOR_SEQ = 20
    OTHER_FEATURES = {'CDS', 'start_codon', 'stop_codon', 'UTR'}
    DEFAULT_RNA_TYPES = {'transcript', 'RNA', 'mRNA', 'miRNA', 'ncRNA', 'tRNA', 'rRNA', 'snRNA', 'lnc_RNA', 'snoRNA',
                         'tRNA', 'telomerase_RNA', 'Y_RNA', 'scRNA'}

    def __init__(self, gene_db_list, db, delta=0, prepare_profiles=True):
        if db is None:
            return

        assert gene_db_list
        # gffutils main structure
        self.db = db
        # list of genes in cluster
        self.gene_db_list = gene_db_list
        # gene region
        self.chr_id, self.start, self.end = self.get_gene_region()
        self.delta = delta

        # additional info for canonical splice site detection
        self.all_read_region_start = self.start
        self.all_read_region_end = self.end
        self.canonical_sites = {}
        self.gene_regions = {}
        self.reference_region = None

        # profiles for all known isoforoms
        self.intron_profiles = FeatureProfiles()
        self.exon_profiles = FeatureProfiles()
        self.split_exon_profiles = FeatureProfiles()

        self.all_isoforms_introns, self.all_isoforms_exons = self.set_introns_and_exons()
        if not self.all_isoforms_exons:
            logger.warning("Genes %s have no exons, check you GTF file" % ", ".join([g.id for g in gene_db_list]))
        if prepare_profiles:
            self.split_exon_profiles.set_features(self.split_exons(self.exon_profiles.features))
            self.set_junction_profiles(self.all_isoforms_introns, self.all_isoforms_exons)

        self.isoform_strands = {}
        self.gene_strands = {}
        self.set_isoform_strands()
        self.other_features = {}
        self.set_other_features()
        self.sources = {}
        self.set_sources()
        self.gene_id_map = {}
        self.set_gene_ids()
        self.feature_attributes = {}
        self.set_gene_attributes()
        if prepare_profiles:
            self.exon_property_map = self.set_feature_properties(self.all_isoforms_exons, self.exon_profiles)
            self.intron_property_map = self.set_feature_properties(self.all_isoforms_introns, self.intron_profiles)

    @classmethod
    def from_models(cls, transcript_model_storage, delta=0):
        gene_info = cls.__new__(cls)
        gene_info.db = None
        gene_info.gene_db_list = []
        if not transcript_model_storage:
            return cls([], None, delta)

        # gene region
        gene_info.chr_id = transcript_model_storage[0].chr_id
        gene_info.start = transcript_model_storage[0].get_start()
        gene_info.end = transcript_model_storage[0].get_end()
        gene_info.delta = delta
        gene_info.all_isoforms_exons = {}
        gene_info.all_isoforms_introns = {}
        gene_info.isoform_strands = {}
        gene_info.sources = {}
        gene_info.other_features = {}
        gene_info.gene_id_map = {}
        gene_info.feature_attributes = {}
        introns = set()
        exons = set()

        for transcript_model in transcript_model_storage:
            gene_info.start = min(gene_info.start, transcript_model.get_start())
            gene_info.end = max(gene_info.end, transcript_model.get_end())
            t_id = transcript_model.transcript_id
            gene_info.all_isoforms_exons[t_id] = transcript_model.exon_blocks
            gene_info.all_isoforms_introns[t_id] = junctions_from_blocks(transcript_model.exon_blocks)
            exons.update(gene_info.all_isoforms_exons[t_id])
            introns.update(gene_info.all_isoforms_introns[t_id])
            gene_info.isoform_strands[transcript_model.transcript_id] = transcript_model.strand
            gene_info.sources[transcript_model.transcript_id] = transcript_model.source
            gene_info.sources[transcript_model.gene_id] = transcript_model.source
            gene_info.other_features[transcript_model.transcript_id] = transcript_model.other_features
            gene_info.gene_id_map[transcript_model.transcript_id] = transcript_model.gene_id

        # profiles for all known isoforoms
        gene_info.intron_profiles = FeatureProfiles()
        gene_info.exon_profiles = FeatureProfiles()
        gene_info.split_exon_profiles = FeatureProfiles()
        gene_info.ambiguous_isoforms = set()

        gene_info.intron_profiles.set_features(sorted(list(introns)))
        gene_info.exon_profiles.set_features(sorted(list(exons)))
        gene_info.split_exon_profiles.set_features(gene_info.split_exons(gene_info.exon_profiles.features))

        for transcript_model in transcript_model_storage:
            transcript_region = (transcript_model.get_start(), transcript_model.get_end())
            t_id = transcript_model.transcript_id
            gene_info.intron_profiles.set_profiles(t_id, gene_info.all_isoforms_introns[t_id], transcript_region, partial(equal_ranges, delta=0))
            gene_info.exon_profiles.set_profiles(t_id, gene_info.all_isoforms_exons[t_id], transcript_region, partial(equal_ranges, delta=0))
            gene_info.split_exon_profiles.set_profiles(t_id, gene_info.all_isoforms_exons[t_id], transcript_region, contains)

        gene_info.regions_for_bam_fetch = [(gene_info.start, gene_info.end)]
        gene_info.exon_property_map = None
        gene_info.intron_property_map = None

        # additional info for canonical splice site detection
        gene_info.all_read_region_start = gene_info.start
        gene_info.all_read_region_end = gene_info.end
        gene_info.canonical_sites = {}
        gene_info.gene_regions = {}
        gene_info.reference_region = None

        return gene_info

    @classmethod
    def from_model(cls, transcript_model, delta=0):
        gene_info = cls.__new__(cls)
        gene_info.db = None
        gene_info.gene_db_list = []
        # gene region
        gene_info.chr_id = transcript_model.chr_id
        gene_info.start = transcript_model.get_start()
        gene_info.end = transcript_model.get_end()
        gene_info.delta = delta

        # profiles for all known isoforoms
        gene_info.intron_profiles = FeatureProfiles()
        gene_info.exon_profiles = FeatureProfiles()
        gene_info.split_exon_profiles = FeatureProfiles()
        gene_info.ambiguous_isoforms = set()

        exons = transcript_model.exon_blocks
        introns = junctions_from_blocks(transcript_model.exon_blocks)
        t_id = transcript_model.transcript_id
        transcript_region = (transcript_model.get_start(), transcript_model.get_end())

        gene_info.all_isoforms_exons = {t_id: exons}
        gene_info.all_isoforms_introns = {t_id: introns}

        gene_info.exon_profiles.set_features(exons)
        gene_info.split_exon_profiles.set_features(exons)
        gene_info.intron_profiles.set_features(introns)

        gene_info.intron_profiles.set_profiles(t_id, introns, transcript_region, partial(equal_ranges, delta=0))
        gene_info.exon_profiles.set_profiles(t_id, exons, transcript_region, partial(equal_ranges, delta=0))
        gene_info.split_exon_profiles.set_profiles(t_id, exons, transcript_region, contains)

        gene_info.isoform_strands = {transcript_model.transcript_id: transcript_model.strand}
        gene_info.sources = {transcript_model.transcript_id: transcript_model.source,
                             transcript_model.gene_id: transcript_model.source}
        gene_info.other_features = {transcript_model.transcript_id: transcript_model.other_features}
        gene_info.gene_id_map = {transcript_model.transcript_id: transcript_model.gene_id}
        gene_info.feature_attributes = {}

        gene_info.regions_for_bam_fetch = [(gene_info.start, gene_info.end)]
        gene_info.exon_property_map = None
        gene_info.intron_property_map = None

        # additional info for canonical splice site detection
        gene_info.all_read_region_start = gene_info.start
        gene_info.all_read_region_end = gene_info.end
        gene_info.canonical_sites = {}
        gene_info.gene_regions = {}
        gene_info.reference_region = None

        return gene_info

    @classmethod
    def from_region(cls, chr_id, start, end, delta=0, chr_record=None):
        gene_info = cls.__new__(cls)
        gene_info.db = None
        gene_info.gene_db_list = []
        # gene region
        gene_info.chr_id = chr_id
        gene_info.start = start
        gene_info.end = end
        gene_info.delta = delta

        # profiles for all known isoforms
        gene_info.intron_profiles = FeatureProfiles()
        gene_info.exon_profiles = FeatureProfiles()
        gene_info.split_exon_profiles = FeatureProfiles()
        gene_info.ambiguous_isoforms = set()

        gene_info.all_isoforms_exons = {}
        gene_info.all_isoforms_introns = {}
        gene_info.isoform_strands = {}
        gene_info.other_features = {}
        gene_info.sources = {}
        gene_info.gene_id_map = {}
        gene_info.feature_attributes = {}
        gene_info.regions_for_bam_fetch = [(start, end)]
        gene_info.exon_property_map = None
        gene_info.intron_property_map = None

        # additional info for canonical splice site detection
        gene_info.all_read_region_start = gene_info.start
        gene_info.all_read_region_end = gene_info.end
        gene_info.canonical_sites = {}
        gene_info.gene_regions = {}
        gene_info.reference_region = None
        if chr_record:
            gene_info.reference_region = str(chr_record[gene_info.all_read_region_start - 1:gene_info.all_read_region_end + 1])

        return gene_info

    @classmethod
    def deserialize(cls, infile, genedb):
        gene_info = cls.__new__(cls)
        gene_info.db = genedb
        gene_info.delta = read_int(infile)
        gene_info.gene_db_list = []

        gene_count = read_int(infile)
        for i in range(gene_count):
            gene_id = read_string(infile)
            if gene_info.db:
                gene_info.gene_db_list.append(gene_info.db[gene_id])
        gene_info.chr_id = read_string(infile)
        gene_info.start = read_int(infile)
        gene_info.end = read_int(infile)

        gene_info.all_read_region_start = gene_info.start
        gene_info.all_read_region_end = gene_info.end

        # the rest is computed based on the database
        gene_info.reference_region = None
        gene_info.canonical_sites = {}
        gene_info.gene_regions = {}
        gene_info.intron_profiles = FeatureProfiles()
        gene_info.exon_profiles = FeatureProfiles()
        gene_info.split_exon_profiles = FeatureProfiles()
        gene_info.all_isoforms_introns = {}
        gene_info.all_isoforms_exons = {}
        if gene_info.gene_db_list:
            gene_info.all_isoforms_introns, gene_info.all_isoforms_exons = gene_info.set_introns_and_exons()
            gene_info.split_exon_profiles.set_features(gene_info.split_exons(gene_info.exon_profiles.features))
            gene_info.set_junction_profiles(gene_info.all_isoforms_introns, gene_info.all_isoforms_exons)

        gene_info.isoform_strands = {}
        gene_info.gene_strands = {}
        gene_info.set_isoform_strands()
        gene_info.other_features = {}
        gene_info.set_other_features()
        gene_info.sources = {}
        gene_info.set_sources()
        gene_info.gene_id_map = {}
        gene_info.set_gene_ids()
        gene_info.feature_attributes = {}
        gene_info.set_gene_attributes()
        gene_info.exon_property_map = gene_info.set_feature_properties(gene_info.all_isoforms_exons, gene_info.exon_profiles)
        gene_info.intron_property_map = gene_info.set_feature_properties(gene_info.all_isoforms_introns, gene_info.intron_profiles)
        return gene_info

    def serialize(self, outfile):
        write_int(self.delta, outfile)
        write_int(len(self.gene_db_list), outfile)
        for g in self.gene_db_list:
            write_string(g.id, outfile)
        write_string(self.chr_id, outfile)
        write_int(self.start, outfile)
        write_int(self.end, outfile)

    def empty(self):
        return not self.exon_profiles.features

    def print_debug(self):
        gene_names = []
        for g in self.gene_db_list:
            gene_names.append(g.id)
        logger.debug(">>> Processing gene set:" + ", ".join(gene_names))
        logger.debug("Split exon profiles:")
        self.split_exon_profiles.print_debug()
        logger.debug("Intron profiles:")
        self.intron_profiles.print_debug()
        logger.debug(">>> End of  gene informaiotn")

    # return region of overlapping gene set
    def get_gene_region(self):
        start = self.gene_db_list[0].start
        end = max(g.end for g in self.gene_db_list)
        chr_id = self.gene_db_list[0].seqid

        for gene_db in self.gene_db_list:
            if start > gene_db.start:
                start = gene_db.start
            if end < gene_db.end:
                end = gene_db.end

        return chr_id, start, end

    # return regions of every gene
    def get_gene_regions(self):
        if self.gene_regions:
            return self.gene_regions
        self.gene_regions = {}
        for gene_db in self.gene_db_list:
            self.gene_regions[gene_db.id] = (gene_db.start, gene_db.end)
        return self.gene_regions

    # set strands
    def set_isoform_strands(self):
        self.isoform_strands = {}
        for gene_db in self.gene_db_list:
            for t in self.db.children(gene_db, featuretype=self.DEFAULT_RNA_TYPES):
                self.isoform_strands[t.id] = t.strand
        self.gene_strands = {}
        for gene_db in self.gene_db_list:
            self.gene_strands[gene_db.id] = gene_db.strand

    def set_other_features(self):
        self.other_features = defaultdict(list)
        for gene in self.gene_db_list:
            for t in self.db.children(gene, featuretype=self.DEFAULT_RNA_TYPES):
                for e in self.db.children(t):
                    if e.featuretype in GeneInfo.OTHER_FEATURES:
                        self.other_features[t.id].append((e.start, e.end, e.featuretype))

    def set_sources(self):
        self.sources = {}
        for gene in self.gene_db_list:
            self.sources[gene.id] = gene.source
            for t in self.db.children(gene, featuretype=self.DEFAULT_RNA_TYPES):
                self.sources[t.id] = t.source

    # set isoform_id -> gene_id map
    def set_gene_ids(self):
        self.gene_id_map = {}
        for gene_db in self.gene_db_list:
            for t in self.db.children(gene_db, featuretype=self.DEFAULT_RNA_TYPES):
                self.gene_id_map[t.id] = gene_db.id

    def set_gene_attributes(self):
        self.feature_attributes = defaultdict(str)
        for gene_db in self.gene_db_list:
            for attr in gene_db.attributes.keys():
                if attr in ['gene_id', 'ID', 'level', 'Parent']:
                    continue
                if gene_db.attributes[attr]:
                    self.feature_attributes[gene_db.id] += '%s "%s"; ' % (attr, gene_db.attributes[attr][0])
            for t in self.db.children(gene_db, featuretype=self.DEFAULT_RNA_TYPES):
                for attr in t.attributes.keys():
                    if attr in ['transcript_id', 'gene_id', 'ID', 'level', 'exons', 'Parent']:
                        continue
                    if t.attributes[attr]:
                        self.feature_attributes[t.id] += '%s "%s"; ' % (attr, t.attributes[attr][0])
                for e in self.db.children(gene_db, featuretype=('exon')):
                    exon_id = t.id + "_%d_%d_%s" % (e.start, e.end, e.strand)
                    for attr in t.attributes.keys():
                        if attr in ['transcript_id', 'gene_id', 'ID', 'Parent', 'level', 'exon_id', 'exon', 'exon_number']:
                            continue
                        if t.attributes[attr]:
                            self.feature_attributes[exon_id] += '%s "%s"; ' % (attr, t.attributes[attr][0])

    # assigns an ordered list of all known exons and introns to self.exons and self.introns
    # returns 2 maps, isoform id -> intron / exon list
    def set_introns_and_exons(self):
        used_cds = False
        # dictionary: isoform id -> ordered list of intron coordinates
        all_isoforms_introns = {}
        # dictionary: isoform id -> ordered list of exon coordinates
        all_isoforms_exons = {}

        for gene_db in self.gene_db_list:
            for t in self.db.children(gene_db, featuretype=self.DEFAULT_RNA_TYPES, order_by='start'):
                exons = []
                all_features = []
                for e in self.db.children(t, featuretype=('exon', 'CDS'), order_by='start'):
                    all_features.append((e.featuretype, e.start, e.end))

                for i, e in enumerate(all_features):
                    if e[0] == 'exon':
                        exons.append((e[1], e[2]))
                        continue

                    cds = (e[1], e[2])
                    prev_exon = None if i == 0 else (all_features[i-1][1], all_features[i-1][2])
                    next_exon = None if i == len(all_features)-1 else (all_features[i+1][1], all_features[i+1][2])
                    if (prev_exon and overlaps(cds, prev_exon)) or (next_exon and overlaps(cds, next_exon)):
                        continue
                    exons.append(cds)
                    used_cds = True

                if not exons:
                    logger.warning("Malformed transcript %s has no exons or CDS" % t.id)
                    continue
                all_isoforms_exons[t.id] = exons
                all_isoforms_introns[t.id] = junctions_from_blocks(all_isoforms_exons[t.id])

        if self.db and not all_isoforms_exons:
            logger.warning("Gene %s has no exons / transcripts, check your input annotation" % self.gene_db_list[0].id)

        if used_cds:
            logger.warning("Gene %s has only CDS features for some transcripts, will be treating those as exons" % self.gene_db_list[0].id)

        introns = set()
        exons = set()
        for i in all_isoforms_introns.keys():
            introns.update(all_isoforms_introns[i])
        for i in all_isoforms_exons.keys():
            exons.update(all_isoforms_exons[i])

        self.intron_profiles.set_features(sorted(list(introns)))
        self.exon_profiles.set_features(sorted(list(exons)))

        return all_isoforms_introns, all_isoforms_exons

    # set feature properties for exon/intron counts
    def set_feature_properties(self, isoforms_to_feature_map, feature_profiles):
        similar_features = set()
        contained_features = set()
        # TODO: change to interval tree instead of brute force
        for f1 in feature_profiles.features:
            for f2 in feature_profiles.features:
                if f1 == f2:
                    continue
                if equal_ranges(f1, f2, self.delta):
                    similar_features.add(f1)
                    similar_features.add(f2)
                if contains(f1, f2):
                    contained_features.add(f2)
                elif contains(f2, f1):
                    contained_features.add(f1)

        feature_to_isoform = defaultdict(list)
        for t in isoforms_to_feature_map.keys():
            isoform_features = isoforms_to_feature_map[t]
            if len(isoform_features) == 0:
                continue
            elif len(isoform_features) == 1:
                feature_to_isoform[isoform_features[0]].append((t, 'T'))
            else:
                feature_to_isoform[isoform_features[0]].append((t, 'T'))
                feature_to_isoform[isoform_features[-1]].append((t, 'T'))
                for e in isoform_features[1:-1]:
                    feature_to_isoform[e].append((t, ''))

        feature_properties = []
        for feature in feature_profiles.features:
            if all(x[1] == 'T' for x in feature_to_isoform[feature]):
                feature_type = "X"
            elif any(x[1] == 'T' for x in feature_to_isoform[feature]):
                feature_type = "T"
            else:
                feature_type = "I"

            if feature in similar_features:
                # similar features
                feature_type += "S"
            if feature in contained_features:
                # feature contained in another one
                feature_type += "C"

            strands = sorted(set([self.isoform_strands[t[0]] for t in feature_to_isoform[feature]]))
            strand_str = "".join(strands)
            gene_ids = set([self.gene_id_map[t[0]] for t in feature_to_isoform[feature]])

            if len(feature_to_isoform[feature]) == 1:
                # unique feature, appears only in one isoform
                feature_type += "U"
            elif len(gene_ids) > 1:
                # multiple genes
                feature_type += "M"

            feature_properties.append(FeatureInfo(self.chr_id, feature[0], feature[1], strand_str,
                                                  feature_type, list(gene_ids)))

        assert len(feature_properties) == len(feature_profiles.features)
        return feature_properties

    # split exons into non-overlapping covering blocks
    @staticmethod
    def split_exons(exons):
        exon_starts = sorted(map(lambda x: x[0], exons))
        exon_ends = sorted(map(lambda x: x[1], exons))

        current_state = 0
        starts_pos = 0
        ends_pos = 0
        last_border = -1
        exon_blocks = []

        while starts_pos < len(exon_starts):
            if exon_starts[starts_pos] <= exon_ends[ends_pos]:
                # do not consider the same exon star
                if starts_pos == 0 or exon_starts[starts_pos] > exon_starts[starts_pos - 1]:
                    cur_border = exon_starts[starts_pos]
                    if last_border != -1 and current_state > 0:
                        exon_blocks.append((last_border, cur_border - 1))
                    last_border = cur_border
                current_state += 1
                starts_pos += 1
            else:
                if last_border == -1 or current_state == 0:
                    print("Error, exon ends before the start")

                if ends_pos == 0 or exon_ends[ends_pos] > exon_ends[ends_pos - 1]:
                    cur_border = exon_ends[ends_pos]
                    exon_blocks.append((last_border, cur_border))
                    last_border = cur_border + 1
                current_state -= 1
                ends_pos += 1

        while ends_pos < len(exon_ends):
            if ends_pos == 0 or exon_ends[ends_pos] > exon_ends[ends_pos - 1]:
                cur_border = exon_ends[ends_pos]
                exon_blocks.append((last_border, cur_border))
                last_border = cur_border + 1
            current_state -= 1
            ends_pos += 1

        if current_state != 0:
            logger.critical("Unequal number of starts and ends")

        if exon_blocks != sorted(exon_blocks):
            logger.critical("Somehow block are unsorted")

        return exon_blocks

    # calculate junction profiles for known isoforms
    def set_junction_profiles(self, all_isoforms_introns, all_isoforms_exons):
        for gene_db in self.gene_db_list:
            for t_id in self.all_isoforms_exons.keys():
                transcript_region = self.transcript_region(t_id)
                # setting up intron profiles for current isoform
                self.intron_profiles.set_profiles(t_id, all_isoforms_introns[t_id], transcript_region,
                                                  partial(equal_ranges, delta=0))
                # setting up exon profiles for current isoform
                self.exon_profiles.set_profiles(t_id, all_isoforms_exons[t_id], transcript_region,
                                                partial(equal_ranges, delta=0))
                # setting up split exon profiles for current isoform
                self.split_exon_profiles.set_profiles(t_id, all_isoforms_exons[t_id], transcript_region, contains)

    def transcript_start(self, transcript_id):
        return self.all_isoforms_exons[transcript_id][0][0]

    def transcript_end(self, transcript_id):
        return self.all_isoforms_exons[transcript_id][-1][1]

    def transcript_region(self, transcript_id):
        exons = self.all_isoforms_exons[transcript_id]
        return exons[0][0], exons[-1][1]

    def transcript_exon_count(self, transcript_id):
        return sum([1 if e == 1 else 0 for e in self.exon_profiles.profiles[transcript_id]])

    def total_transcript_length(self, transcript_id):
        return intervals_total_length(self.all_isoforms_exons[transcript_id])

    # closed interval
    def get_ref_seq(self, ref_start, ref_end):
        assert self.reference_region
        if ref_start < self.all_read_region_start or ref_end > self.all_read_region_end:
            logger.warning("Trying to extract ref seq outside of reference region: %d:%d" % (ref_start, ref_end))
        left_pos = ref_start - self.all_read_region_start
        right_pos = ref_end - self.all_read_region_start
        return self.reference_region[left_pos:right_pos+1]

    def set_reference_sequence(self, start, end, chr_record):
        self.all_read_region_start = start
        self.all_read_region_end = end
        self.reference_region = \
            str(chr_record[self.all_read_region_start - 1:self.all_read_region_end])
        self.canonical_sites = {}


class StrandDetector:
    def __init__(self, chr_record):
        self.strand_dict = {}
        self.chr_record = chr_record

    def set_strand(self, intron, strand=None):
        if strand:
            self.strand_dict[intron] = strand
        elif self.chr_record:
            self.strand_dict[intron] = get_intron_strand(intron, self.chr_record)

    def count_canonical_sites(self, introns):
        count_fwd = 0
        count_rev = 0
        for intron in introns:
            if not intron in self.strand_dict:
                strand = get_intron_strand(intron, self.chr_record)
                self.strand_dict[intron] = strand
            else:
                strand = self.strand_dict[intron]
            if strand == '+':
                count_fwd += 1
            elif strand == '-':
                count_rev += 1
        return count_fwd, count_rev

    # all splice sites must be canonical from the same strand, not just the majority
    def get_clean_strand(self, introns):
        count_fwd, count_rev = self.count_canonical_sites(introns)
        if count_fwd == 0 and count_rev > 0:
            return  '-'
        elif count_fwd > 0 and count_rev == 0:
            return '+'
        return '.'

    def get_strand(self, introns, has_polya=False, has_polyt=False):
        count_fwd, count_rev = self.count_canonical_sites(introns)
        if count_fwd == count_rev:
            if has_polya and not has_polyt:
                return '+'
            elif has_polyt and not has_polya:
                return '-'
            return '.'
        return '+' if count_rev < count_fwd else '-'


def get_all_chromosome_genes(genedb, chr_id):
    return [g.id for g in genedb.region(seqid=chr_id, start=1, featuretype="gene")]


def get_all_chromosome_transcripts(genedb, chr_id):
    return [g.id for g in genedb.region(seqid=chr_id, start=1, featuretype=GeneInfo.DEFAULT_RNA_TYPES)]


############################################################################
# Copyright (c) 2023-2026 University of Helsinki
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

"""
Barcode calling for single-cell and spatial transcriptomics.

This package provides k-mer indexing and barcode detection for various platforms:
- Curio (double barcode)
- Stereo-seq (standard and splitting modes)
- 10x Genomics (v3, Visium HD)

Subpackages:
    indexers: K-mer indexing implementations
    callers: Barcode detector implementations
"""

# Indexers
from .indexers import (
    KmerIndexer,
    ArrayKmerIndexer,
    Dict2BitKmerIndexer,
    Array2BitKmerIndexer,
    SharedMemoryArray2BitKmerIndexer,
    SharedMemoryIndexInfo,
)

# Result classes
from .callers import (
    BarcodeDetectionResult,
    LinkerBarcodeDetectionResult,
    TSOBarcodeDetectionResult,
    TenXBarcodeDetectionResult,
    SplittingBarcodeDetectionResult,
    ReadStats,
    increase_if_valid,
)

# Detector classes
from .callers import (
    CurioBarcodeDetector,
    CurioIlluminaDetector,
    StereoBarcodeDetector,
    SharedMemoryStereoBarcodeDetector,
    StereoSplittingBarcodeDetector,
    SharedMemoryStereoSplittingBarcodeDetector,
    SharedMemoryWrapper,
    TenXBarcodeDetector,
    TenXv2BarcodeDetector,
    VisiumHDBarcodeDetector,
    UniversalSingleMoleculeExtractor,
    MoleculeStructure
)

# Utilities
from .common import str_to_2bit, bit_to_str, find_polyt_start, batch_str_to_2bit, batch_str_to_2bit_chunked

# Barcode detection pipeline functions
from .detect_barcodes import (
    process_single_thread,
    process_in_parallel,
    get_umi_length,
    create_barcode_caller,
    BARCODE_CALLING_MODES,
    BARCODE_FILES_REQUIRED,
)

__all__ = [
    # Indexers
    'KmerIndexer',
    'ArrayKmerIndexer',
    'Dict2BitKmerIndexer',
    'Array2BitKmerIndexer',
    'SharedMemoryArray2BitKmerIndexer',
    'SharedMemoryIndexInfo',
    # Result classes
    'BarcodeDetectionResult',
    'LinkerBarcodeDetectionResult',
    'TSOBarcodeDetectionResult',
    'TenXBarcodeDetectionResult',
    'SplittingBarcodeDetectionResult',
    'ReadStats',
    'increase_if_valid',
    # Curio detectors
    'CurioBarcodeDetector',
    'CurioIlluminaDetector',
    # Stereo detectors
    'StereoBarcodeDetector',
    'SharedMemoryStereoBarcodeDetector',
    'StereoSplittingBarcodeDetector',
    'SharedMemoryStereoSplittingBarcodeDetector',
    'SharedMemoryWrapper',
    # 10x detectors
    'TenXBarcodeDetector',
    'TenXv2BarcodeDetector',
    'VisiumHDBarcodeDetector',
    # universal calling
    'UniversalSingleMoleculeExtractor',
    'MoleculeStructure',
    # Utilities
    'str_to_2bit',
    'bit_to_str',
    'find_polyt_start',
    'batch_str_to_2bit',
    'batch_str_to_2bit_chunked',
    # Detection pipeline
    'process_single_thread',
    'process_in_parallel',
    'get_umi_length',
    'create_barcode_caller',
    'BARCODE_CALLING_MODES',
    'BARCODE_FILES_REQUIRED',
]



############################################################################
# Copyright (c) 2025-2026 University of Helsinki
# All Rights Reserved
# See file LICENSE for details.
############################################################################

import pytest
from src.barcode_calling.common import (
    str_to_2bit,
    bit_to_str,
    batch_str_to_2bit,
    batch_str_to_2bit_chunked,
    reverese_complement,
    find_polyt_start,
    align_pattern_ssw,
    find_candidate_with_max_score_ssw,
    find_candidate_with_max_score_ssw_var_len,
    detect_exact_positions,
    detect_first_exact_positions,
    NUCL2BIN,
    BIN2NUCL,
    base_comp,
)


class TestBitEncoding:
    """Test 2-bit DNA encoding functions."""

    def test_str_to_2bit_single_nucleotides(self):
        """Test encoding single nucleotides."""
        # Based on implementation: A=0, C=1, T=2, G=3
        assert str_to_2bit("A") == 0
        assert str_to_2bit("C") == 1
        assert str_to_2bit("T") == 2
        assert str_to_2bit("G") == 3

    def test_str_to_2bit_lowercase(self):
        """Test that lowercase works the same."""
        assert str_to_2bit("a") == str_to_2bit("A")
        assert str_to_2bit("c") == str_to_2bit("C")
        assert str_to_2bit("t") == str_to_2bit("T")
        assert str_to_2bit("g") == str_to_2bit("G")

    def test_str_to_2bit_sequence(self):
        """Test encoding multi-nucleotide sequences."""
        # "AC" = A(00) C(01) = 0b0001 = 1
        assert str_to_2bit("AC") == 1
        # "CA" = C(01) A(00) = 0b0100 = 4
        assert str_to_2bit("CA") == 4
        # "TG" = T(10) G(11) = 0b1011 = 11
        assert str_to_2bit("TG") == 11

    def test_bit_to_str_single_nucleotides(self):
        """Test decoding single nucleotides."""
        assert bit_to_str(0, 1) == "A"
        assert bit_to_str(1, 1) == "C"
        assert bit_to_str(2, 1) == "T"
        assert bit_to_str(3, 1) == "G"

    def test_bit_to_str_sequence(self):
        """Test decoding multi-nucleotide sequences."""
        assert bit_to_str(1, 2) == "AC"
        assert bit_to_str(4, 2) == "CA"
        assert bit_to_str(11, 2) == "TG"

    def test_roundtrip_encoding(self):
        """Test that encoding and decoding are inverses."""
        sequences = ["A", "ACTG", "GGGG", "AAAA", "TTTT", "ACGTACGT",
                     "ACTGACTGACTGACTG", "TATATATA"]
        for seq in sequences:
            encoded = str_to_2bit(seq)
            decoded = bit_to_str(encoded, len(seq))
            assert decoded == seq, f"Failed for {seq}"

    def test_encoding_constants(self):
        """Test that NUCL2BIN and BIN2NUCL are consistent."""
        assert BIN2NUCL[NUCL2BIN['A']] == 'A'
        assert BIN2NUCL[NUCL2BIN['C']] == 'C'
        assert BIN2NUCL[NUCL2BIN['T']] == 'T'
        assert BIN2NUCL[NUCL2BIN['G']] == 'G'


class TestReverseComplement:
    """Test reverse complement function."""

    def test_single_nucleotides(self):
        """Test complementing single nucleotides."""
        assert reverese_complement("A") == "T"
        assert reverese_complement("T") == "A"
        assert reverese_complement("C") == "G"
        assert reverese_complement("G") == "C"

    def test_sequence(self):
        """Test reverse complement of sequences."""
        assert reverese_complement("ACTG") == "CAGT"
        assert reverese_complement("AAAA") == "TTTT"
        assert reverese_complement("CCCC") == "GGGG"

    def test_palindrome(self):
        """Test that palindromic sequences return themselves."""
        # GAATTC is EcoRI recognition site (palindrome)
        assert reverese_complement("GAATTC") == "GAATTC"
        # ATAT is also palindromic
        assert reverese_complement("ATAT") == "ATAT"

    def test_with_n(self):
        """Test handling of N (unknown) nucleotide."""
        assert reverese_complement("N") == "N"
        assert reverese_complement("ACNG") == "CNGT"
        assert reverese_complement("NNN") == "NNN"

    def test_base_comp_dict(self):
        """Test base_comp dictionary is complete."""
        assert base_comp['A'] == 'T'
        assert base_comp['T'] == 'A'
        assert base_comp['C'] == 'G'
        assert base_comp['G'] == 'C'
        assert base_comp['N'] == 'N'


class TestFindPolytStart:
    """Test polyT detection function."""

    def test_clear_polyt_tract(self):
        """Test finding clear polyT tract."""
        # 10 bases of random + 16 Ts
        seq = "ACGTACGTAC" + "T" * 16
        result = find_polyt_start(seq)
        assert result >= 10  # Should find polyT starting around position 10

    def test_no_polyt(self):
        """Test sequence without polyT."""
        seq = "ACGTACGTACGTACGTACGT"
        assert find_polyt_start(seq) == -1

    def test_short_sequence(self):
        """Test sequence shorter than window size."""
        seq = "ACTG"  # Shorter than default window_size=16
        assert find_polyt_start(seq) == -1

    def test_polyt_at_start(self):
        """Test polyT at the beginning."""
        seq = "T" * 20 + "ACGT"
        result = find_polyt_start(seq)
        assert result >= 0 and result < 5  # Should be near start

    def test_mixed_with_some_t(self):
        """Test sequence with some Ts but not enough for polyT."""
        seq = "ATATATAT" * 4  # Alternating, not enough consecutive Ts
        assert find_polyt_start(seq) == -1

    def test_polyt_with_interruption(self):
        """Test polyT with slight interruption still detected."""
        # 12 Ts, then A, then 4 more Ts - should still be >= 75% T
        seq = "ACGTACGT" + "TTTTTTTTTTTTATTTTT"
        result = find_polyt_start(seq)
        assert result >= 0  # Should still detect


class TestSSWAlignment:
    """Test SSW-based alignment functions."""

    def test_align_pattern_exact_match(self):
        """Test aligning exact match pattern."""
        seq = "NNNNACTGACTGNNNN"
        start, end, pstart, pend, score = align_pattern_ssw(
            seq, 0, len(seq), "ACTGACTG", min_score=8)
        assert score == 8  # Perfect match of 8 bases
        assert start is not None

    def test_align_pattern_no_match(self):
        """Test aligning with no matching pattern."""
        seq = "AAAAAAAAAAAAAAAA"
        result = align_pattern_ssw(seq, 0, len(seq), "GGGGGGGG", min_score=6)
        assert result[0] is None  # No match above threshold

    def test_align_pattern_partial_match(self):
        """Test aligning with partial match."""
        seq = "NNNNACTGNNNN"
        start, end, pstart, pend, score = align_pattern_ssw(
            seq, 0, len(seq), "ACTG", min_score=3)
        assert score >= 3
        assert start is not None

    def test_align_pattern_with_mismatch(self):
        """Test aligning pattern with mismatch."""
        seq = "NNNNACTAACTGNNNN"  # ACTA instead of ACTG
        start, end, pstart, pend, score = align_pattern_ssw(
            seq, 0, len(seq), "ACTG", min_score=3)
        assert score >= 3  # Should still find partial match


class TestFindCandidateSSW:
    """Test barcode candidate selection functions."""

    def test_find_candidate_exact_match(self):
        """Test finding exact barcode match."""
        barcode_matches = [("ACTGACTG", 5, [0, 1, 2])]
        read_seq = "NNACTGACTGNN"

        barcode, score, start, end = find_candidate_with_max_score_ssw(
            barcode_matches, read_seq, min_score=6)

        assert barcode == "ACTGACTG"
        assert score >= 6

    def test_find_candidate_multiple_options(self):
        """Test selecting best from multiple candidates."""
        barcode_matches = [
            ("ACTGACTG", 5, [0, 1, 2]),
            ("TGCATGCA", 3, [0, 1]),
        ]
        read_seq = "NNACTGACTGNN"

        barcode, score, start, end = find_candidate_with_max_score_ssw(
            barcode_matches, read_seq, min_score=4)

        assert barcode == "ACTGACTG"  # Should pick the better match

    def test_find_candidate_no_match(self):
        """Test when no candidate meets threshold."""
        barcode_matches = [("ACTGACTG", 5, [0, 1, 2])]
        read_seq = "GGGGGGGGGGGG"

        barcode, score, start, end = find_candidate_with_max_score_ssw(
            barcode_matches, read_seq, min_score=6)

        assert barcode is None

    def test_find_candidate_score_diff(self):
        """Test score difference threshold."""
        # Two similar barcodes
        barcode_matches = [
            ("ACTGACTG", 5, [0, 1, 2]),
            ("ACTGACTA", 5, [0, 1, 2]),
        ]
        read_seq = "NNACTGACTGNN"

        # With high score_diff, should reject ambiguous
        barcode, score, start, end = find_candidate_with_max_score_ssw(
            barcode_matches, read_seq, min_score=4, score_diff=3)

        # Result depends on score difference

    def test_find_candidate_var_len(self):
        """Test variable length barcode matching."""
        barcode_matches = [
            ("ACTGACTGACTG", 5, [0, 1, 2]),  # 12-mer
            ("ACTGACTG", 5, [0, 1, 2]),  # 8-mer
        ]
        read_seq = "NNACTGACTGACTGNN"

        barcode, score, start, end = find_candidate_with_max_score_ssw_var_len(
            barcode_matches, read_seq, min_score=8)

        assert barcode == "ACTGACTGACTG"  # Should prefer longer exact match


class TestNumba2BitEncoding:
    """Test Numba JIT-compiled 2-bit encoding consistency."""

    def test_numba_consistency_with_numpy(self):
        """Test that Numba and NumPy fallback produce identical results."""
        import numpy as np
        from src.barcode_calling.common import NUMBA_AVAILABLE, _convert_chunk_to_2bit

        barcodes = [
            "ACTGACTGACTGACTGACTGACTGA",
            "TGCATGCATGCATGCATGCATGCAT",
            "GGGGGGGGGGGGGGGGGGGGGGGGG",
            "AAAAAAAAAAAAAAAAAAAAAAAAA",
            "CCCCCCCCCCCCCCCCCCCCCCCCC",
        ]
        seq_len = 25

        # Get result from current implementation (Numba if available)
        result = _convert_chunk_to_2bit(barcodes, seq_len)

        # Verify against individual str_to_2bit calls
        for i, bc in enumerate(barcodes):
            expected = str_to_2bit(bc)
            assert result[i] == expected, f"Mismatch for {bc}: got {result[i]}, expected {expected}"

    def test_numba_available_flag(self):
        """Test that NUMBA_AVAILABLE flag is set correctly."""
        from src.barcode_calling.common import NUMBA_AVAILABLE

        # Just verify the flag is a boolean
        assert isinstance(NUMBA_AVAILABLE, bool)

        # If numba is importable without errors, flag should be True
        try:
            from numba import njit, prange
            assert NUMBA_AVAILABLE is True
        except (ImportError, SystemError):
            # SystemError can occur with broken numba installations
            assert NUMBA_AVAILABLE is False


class TestBatchStrTo2Bit:
    """Test batch 2-bit DNA encoding functions."""

    def test_empty_input(self):
        """Test empty input returns empty array."""
        import numpy as np
        result = batch_str_to_2bit([], 25)
        assert len(result) == 0
        assert result.dtype == np.uint64

    def test_single_barcode(self):
        """Test single barcode matches str_to_2bit."""
        barcode = "ACTGACTG"
        result = batch_str_to_2bit([barcode], seq_len=8)
        expected = str_to_2bit(barcode)
        assert result[0] == expected

    def test_multiple_barcodes(self):
        """Test multiple barcodes match individual str_to_2bit calls."""
        barcodes = ["ACTGACTG", "TGCATGCA", "GGGGGGGG", "AAAAAAAA"]
        result = batch_str_to_2bit(barcodes, seq_len=8)

        assert len(result) == 4
        for i, bc in enumerate(barcodes):
            assert result[i] == str_to_2bit(bc), f"Mismatch for barcode {bc}"

    def test_25mer_barcodes(self):
        """Test with default 25-mer barcodes (Stereo-seq length)."""
        barcodes = [
            "ACTGACTGACTGACTGACTGACTGA",
            "TGCATGCATGCATGCATGCATGCAT",
            "GGGGGGGGGGGGGGGGGGGGGGGGG",
        ]
        result = batch_str_to_2bit(barcodes, seq_len=25)

        assert len(result) == 3
        for i, bc in enumerate(barcodes):
            assert result[i] == str_to_2bit(bc), f"Mismatch for barcode {bc}"

    def test_iterator_input(self):
        """Test that iterator input works correctly."""
        barcodes = ["ACTGACTG", "TGCATGCA", "GGGGGGGG"]
        result = batch_str_to_2bit(iter(barcodes), seq_len=8)

        assert len(result) == 3
        for i, bc in enumerate(barcodes):
            assert result[i] == str_to_2bit(bc)

    def test_roundtrip(self):
        """Test that batch encoding can be decoded correctly."""
        barcodes = ["ACTGACTG", "TGCATGCA", "GGGGGGGG"]
        encoded = batch_str_to_2bit(barcodes, seq_len=8)

        for i, bc in enumerate(barcodes):
            decoded = bit_to_str(encoded[i], 8)
            assert decoded == bc, f"Roundtrip failed for {bc}: got {decoded}"


class TestBatchStrTo2BitChunked:
    """Test chunked batch 2-bit DNA encoding function."""

    def test_empty_input(self):
        """Test empty iterator returns empty array."""
        import numpy as np
        result = batch_str_to_2bit_chunked(iter([]), 25)
        assert len(result) == 0
        assert result.dtype == np.uint64

    def test_single_chunk(self):
        """Test processing fewer items than chunk size."""
        barcodes = ["ACTGACTG", "TGCATGCA", "GGGGGGGG"]
        result = batch_str_to_2bit_chunked(iter(barcodes), seq_len=8, chunk_size=100)

        assert len(result) == 3
        for i, bc in enumerate(barcodes):
            assert result[i] == str_to_2bit(bc)

    def test_multiple_chunks(self):
        """Test processing more items than chunk size."""
        barcodes = ["ACTGACTG", "TGCATGCA", "GGGGGGGG", "AAAAAAAA", "CCCCCCCC"]
        # Use chunk_size=2 to force multiple chunks
        result = batch_str_to_2bit_chunked(iter(barcodes), seq_len=8, chunk_size=2)

        assert len(result) == 5
        for i, bc in enumerate(barcodes):
            assert result[i] == str_to_2bit(bc), f"Mismatch for barcode {bc}"

    def test_exact_chunk_boundary(self):
        """Test when barcode count equals chunk size."""
        barcodes = ["ACTGACTG", "TGCATGCA"]
        result = batch_str_to_2bit_chunked(iter(barcodes), seq_len=8, chunk_size=2)

        assert len(result) == 2
        for i, bc in enumerate(barcodes):
            assert result[i] == str_to_2bit(bc)

    def test_consistency_with_batch(self):
        """Test that chunked and non-chunked produce same results."""
        barcodes = ["ACTGACTG", "TGCATGCA", "GGGGGGGG", "AAAAAAAA", "CCCCCCCC"]

        batch_result = batch_str_to_2bit(barcodes, seq_len=8)
        chunked_result = batch_str_to_2bit_chunked(iter(barcodes), seq_len=8, chunk_size=2)

        assert len(batch_result) == len(chunked_result)
        for i in range(len(batch_result)):
            assert batch_result[i] == chunked_result[i], f"Mismatch at index {i}"


class TestDetectExactPositions:
    """Test detect_exact_positions function."""

    def test_exact_match_in_sequence(self):
        """Test finding exact pattern in sequence."""
        sequence = "NNNNACTGACTGNNNN"
        pattern = "ACTGACTG"
        # Simulate k-mer occurrences: (pattern, count, [positions])
        # Positions are k-mer match positions within the search window
        pattern_occurrences = [(pattern, 5, [4, 5, 6, 7])]

        start, end = detect_exact_positions(
            sequence, 0, len(sequence), kmer_size=4,
            pattern=pattern, pattern_occurrences=pattern_occurrences,
            min_score=6
        )

        assert start is not None
        assert end is not None
        # Should find the pattern starting around position 4
        assert sequence[start:end] == pattern or pattern in sequence[start:end+1]

    def test_pattern_not_in_occurrences(self):
        """Test when pattern is not in the occurrence list."""
        sequence = "NNNNACTGACTGNNNN"
        pattern = "ACTGACTG"
        pattern_occurrences = [("TGCATGCA", 5, [0, 1, 2])]  # Different pattern

        start, end = detect_exact_positions(
            sequence, 0, len(sequence), kmer_size=4,
            pattern=pattern, pattern_occurrences=pattern_occurrences,
            min_score=6
        )

        assert start is None
        assert end is None

    def test_empty_occurrences(self):
        """Test with empty occurrence list."""
        sequence = "NNNNACTGACTGNNNN"

        start, end = detect_exact_positions(
            sequence, 0, len(sequence), kmer_size=4,
            pattern="ACTGACTG", pattern_occurrences=[],
            min_score=6
        )

        assert start is None
        assert end is None

    def test_no_positions_for_pattern(self):
        """Test when pattern has no match positions."""
        sequence = "NNNNACTGACTGNNNN"
        pattern = "ACTGACTG"
        pattern_occurrences = [(pattern, 0, [])]  # No positions

        start, end = detect_exact_positions(
            sequence, 0, len(sequence), kmer_size=4,
            pattern=pattern, pattern_occurrences=pattern_occurrences,
            min_score=6
        )

        assert start is None
        assert end is None


class TestDetectFirstExactPositions:
    """Test detect_first_exact_positions function."""

    def test_finds_first_match(self):
        """Test that first match is returned."""
        sequence = "NNACTGACTGNNACTGACTGNN"
        pattern = "ACTGACTG"
        # Two potential matches - should return first one
        pattern_occurrences = [(pattern, 5, [2, 12])]

        start, end = detect_first_exact_positions(
            sequence, 0, len(sequence), kmer_size=4,
            pattern=pattern, pattern_occurrences=pattern_occurrences,
            min_score=6
        )

        assert start is not None
        # First match should be around position 2
        assert start <= 5

    def test_pattern_not_found(self):
        """Test when pattern is not in occurrence list."""
        sequence = "NNNNACTGACTGNNNN"
        pattern_occurrences = [("TGCATGCA", 5, [0, 1])]

        start, end = detect_first_exact_positions(
            sequence, 0, len(sequence), kmer_size=4,
            pattern="ACTGACTG", pattern_occurrences=pattern_occurrences,
            min_score=6
        )

        assert start is None
        assert end is None

    def test_empty_positions(self):
        """Test with empty position list."""
        sequence = "NNNNACTGACTGNNNN"
        pattern = "ACTGACTG"
        pattern_occurrences = [(pattern, 0, [])]

        start, end = detect_first_exact_positions(
            sequence, 0, len(sequence), kmer_size=4,
            pattern=pattern, pattern_occurrences=pattern_occurrences,
            min_score=6
        )

        assert start is None
        assert end is None

    def test_min_score_threshold(self):
        """Test that min_score filters weak matches."""
        sequence = "NNNNACNNNNNN"  # Only partial match
        pattern = "ACTGACTG"
        pattern_occurrences = [(pattern, 2, [4])]

        # With high min_score, should not find match
        start, end = detect_first_exact_positions(
            sequence, 0, len(sequence), kmer_size=4,
            pattern=pattern, pattern_occurrences=pattern_occurrences,
            min_score=7
        )

        assert start is None
        assert end is None


if __name__ == '__main__':
    pytest.main([__file__, '-v'])



############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# Copyright (c) 2020-2022 Saint Petersburg State University
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import logging
from functools import partial
from collections import defaultdict

from .common import (
    contains,
    equal_ranges,
    junctions_from_blocks,
    left_of,
    overlaps,
    overlaps_at_least,
    overlaps_at_least_when_overlap,
    interval_bin_search,
    interval_bin_search_rev
)

logger = logging.getLogger('IsoQuant')


class MappedReadProfile:
    def __init__(self, gene_profile, read_profile, read_features, gene_profile_range=None):
        self.gene_profile = gene_profile
        self.read_profile = read_profile
        self.read_features = read_features
        # profile range points to indices in the mapped region
        self.gene_profile_range = (0, len(self.gene_profile)) if gene_profile_range is None else gene_profile_range


class CombinedReadProfiles:
    def __init__(self, read_intron_profile, read_exon_profile, read_split_exon_profile,
                 polya_info=None, cage_hits=-1, alignment=None):
        self.read_intron_profile = read_intron_profile
        self.read_exon_profile = read_exon_profile
        self.read_split_exon_profile = read_split_exon_profile
        self.alignment = alignment
        self.polya_info = polya_info
        self.cage_hits = cage_hits


# The following 2 classes are very similar, but lets keep them separately for now
# accepts sorted gapless alignment blocks
class OverlappingFeaturesProfileConstructor:
    # ignore_terminal -- bool flag, indicates whether to ignore leading and trailing -1s in the profile
    def __init__(self, known_features, gene_region,
                 comparator = partial(equal_ranges, delta=0),
                 absence_condition = contains,
                 delta=0):
        self.known_features = known_features
        self.gene_region = gene_region
        self.comparator = comparator
        self.absence_condition = absence_condition
        self.delta = delta

    def construct_intron_profile(self, sorted_blocks, polya_position=-1, polyt_position=-1):
        mapped_region = (sorted_blocks[0][0], sorted_blocks[-1][1])
        read_introns = junctions_from_blocks(sorted_blocks)
        return self.construct_profile_for_features(read_introns, mapped_region, polya_position, polyt_position)

    def construct_exon_profile(self, sorted_blocks, polya_position=-1, polyt_position=-1):
        mapped_region = (sorted_blocks[0][1] + self.delta, sorted_blocks[-1][0] - self.delta)
        return self.construct_profile_for_features(sorted_blocks, mapped_region, polya_position, polyt_position)

    def match_delta(self, feature1, feature2):
        return abs(feature1[0] - feature2[0]) + abs(feature1[1] - feature2[1])

    def match_genomic_features(self, read_features):
        matched_features = defaultdict(list)

        # TODO: starting value can be detected using binary search for long profiles
        gene_pos = 0
        read_pos = 0
        while gene_pos < len(self.known_features) and read_pos < len(read_features):
            if self.comparator(read_features[read_pos], self.known_features[gene_pos]):
                matched_features[read_pos].append(gene_pos)
                gene_pos += 1
            elif overlaps(read_features[read_pos], self.known_features[gene_pos]):
                gene_pos += 1
            elif left_of(read_features[read_pos], self.known_features[gene_pos]):
                read_pos += 1
            else:
                gene_pos += 1
        # eliminating non unique features
        for read_pos in matched_features.keys():
            if len(matched_features[read_pos]) > 1:
                deltas = [self.match_delta(read_features[read_pos], self.known_features[gene_pos])
                          for gene_pos in matched_features[read_pos]]
                best_match = min(deltas)
                filtered_features = []
                for i, d in enumerate(deltas):
                    if d == best_match:
                        filtered_features.append(matched_features[read_pos][i])
                matched_features[read_pos] = filtered_features

        corrected_features = []
        for i, intron in enumerate(read_features):
            if i in matched_features:
                # known feature
                # corresponding known feature, always take first for now
                corrected_features.append(self.known_features[matched_features[i][0]])
            else:
                corrected_features.append(intron)
        return corrected_features

    def construct_profile_for_features(self, read_features, mapped_region=(0, 0), polya_position=-1, polyt_position=-1):
        read_profile = [0] * (len(read_features))
        intron_profile = [0] * (len(self.known_features))
        matched_features = defaultdict(list)

        for i in range(len(intron_profile)):
            if self.absence_condition(mapped_region, self.known_features[i]):
                intron_profile[i] = -1
        for i in range(len(read_profile)):
            if self.absence_condition(self.gene_region, read_features[i]):
                read_profile[i] = -1

        gene_pos = 0
        read_pos = 0
        while gene_pos < len(self.known_features) and read_pos < len(read_features):
            read_feature = read_features[read_pos]
            isoform_feature = self.known_features[gene_pos]
            if read_feature[1] < isoform_feature[0]:
                if read_profile[read_pos] == 0 and gene_pos > 0:
                    read_profile[read_pos] = -1
                read_pos += 1
            elif isoform_feature[1] < read_feature[0]:
                if read_pos > 0:
                    intron_profile[gene_pos] = -1
                gene_pos += 1
            elif self.comparator(read_feature, isoform_feature):
                intron_profile[gene_pos] = 1
                read_profile[read_pos] = 1
                matched_features[read_pos].append(gene_pos)
                gene_pos += 1
            elif overlaps(read_feature, isoform_feature):
                if self.absence_condition(mapped_region, isoform_feature):
                    intron_profile[gene_pos] = -1
                gene_pos += 1

        # eliminating non unique features
        for read_pos in matched_features.keys():
            if len(matched_features[read_pos]) > 1:
                deltas = [self.match_delta(read_features[read_pos], self.known_features[gene_pos])
                          for gene_pos in matched_features[read_pos]]
                best_match = min(deltas)
                for i in range(len(matched_features[read_pos])):
                    if deltas[i] > best_match:
                        intron_profile[matched_features[read_pos][i]] = -1

        corrected_features = []
        for i, v in enumerate(read_profile):
            if v == 1:
                # known feature
                # corresponding known feature, always take first for now
                corrected_features.append(self.known_features[matched_features[i][0]])
            else:
                corrected_features.append(read_features[i])

        # making everying beyond polyA tail as outside feature
        if polya_position != -1:
            for i in range(len(self.known_features)):
                # feature is surely beyond polyA tail
                if self.known_features[i][0] > polya_position + self.delta:
                    intron_profile[i] = -2
        if polyt_position != -1:
            for i in range(len(self.known_features)):
                # feature is surely before polyT tail
                if self.known_features[i][1] < polyt_position - self.delta:
                    intron_profile[i] = -2

        start_pos = 0
        while start_pos < len(intron_profile) and intron_profile[start_pos] == 0:
            start_pos += 1
        end_pos = len(intron_profile) - 1
        while end_pos >= 0 and intron_profile[end_pos] == 0:
            end_pos -= 1

        return MappedReadProfile(intron_profile, read_profile, read_features, (start_pos, end_pos + 1))


# accepts sorted gapless alignment blocks
class NonOverlappingFeaturesProfileConstructor:
    def __init__(self, known_exons, comparator=overlaps, delta=0):
        self.known_exons = known_exons
        self.comparator = comparator
        self.delta = delta

    def construct_profile(self, sorted_blocks, polya_position=-1, polyt_position=-1):
        exon_profile = [0] * (len(self.known_exons))
        read_profile = [0] * (len(sorted_blocks))
        read_exons = sorted_blocks
        gene_pos = 0
        read_pos = 0

        while gene_pos < len(self.known_exons) and read_pos < len(read_exons):
            read_exon = read_exons[read_pos]
            gene_exon = self.known_exons[gene_pos]
            if read_exon[1] < gene_exon[0]:
                if gene_pos > 0 and read_profile[read_pos] == 0:
                    read_profile[read_pos] = -1
                read_pos += 1
            elif gene_exon[1] < read_exon[0]:
                if read_pos > 0 and exon_profile[gene_pos] == 0:
                    exon_profile[gene_pos] = -1
                gene_pos += 1
            elif self.comparator(read_exon, gene_exon):
                exon_profile[gene_pos] = 1
                read_profile[read_pos] = 1
                if read_exon[1] < gene_exon[1]:
                    read_pos += 1
                else:
                    gene_pos += 1
            else:
                if read_exon[1] < gene_exon[1]:
                    read_pos += 1
                else:
                    gene_pos += 1

        # making everything beyond polyA tail as outside feature
        if polya_position != -1:
            polya_index = interval_bin_search(self.known_exons, polya_position + self.delta)
            if polya_index != -1:
                for i in range(polya_index + 1, len(self.known_exons)):
                    exon_profile[i] = -2
        if polyt_position != -1:
            polyt_index = interval_bin_search_rev(self.known_exons, polyt_position - self.delta)
            if polyt_index != -1:
                for i in range(0, polyt_index):
                    exon_profile[i] = -2

        start_pos = 0
        while start_pos < len(exon_profile) and exon_profile[start_pos] == 0:
            start_pos += 1
        end_pos = len(exon_profile) - 1
        while end_pos >= 0 and exon_profile[end_pos] == 0:
            end_pos -= 1

        return MappedReadProfile(exon_profile, read_profile, read_exons, (start_pos, end_pos + 1))


class CombinedProfileConstructor:
    def __init__(self, gene_info, params):
        self.gene_info = gene_info
        self.params = params

        gene_region = (gene_info.start, gene_info.end)
        self.intron_profile_constructor = \
            OverlappingFeaturesProfileConstructor(self.gene_info.intron_profiles.features, gene_region,
                                                  comparator=partial(equal_ranges, delta=self.params.delta),
                                                  absence_condition=partial(overlaps_at_least, delta=self.params.minimal_intron_absence_overlap),
                                                  delta=self.params.delta)
        self.exon_profile_constructor = \
            OverlappingFeaturesProfileConstructor(self.gene_info.exon_profiles.features, gene_region,
                                                  comparator=partial(equal_ranges, delta=self.params.delta),
                                                  delta=self.params.delta)
        self.split_exon_profile_constructor = \
            NonOverlappingFeaturesProfileConstructor(self.gene_info.split_exon_profiles.features,
                                                     comparator=partial(overlaps_at_least_when_overlap,
                                                                        delta=self.params.minimal_exon_overlap),
                                                     delta=self.params.delta)

    def construct_profiles(self, sorted_blocks, polya_info, cage_hits):
        intron_profile = self.intron_profile_constructor.construct_intron_profile(sorted_blocks,
                                                                                  polya_info.external_polya_pos,
                                                                                  polya_info.external_polyt_pos)
        exon_profile = None
        if self.params.count_exons:
            exon_profile = self.exon_profile_constructor.construct_exon_profile(sorted_blocks,
                                                                                polya_info.external_polya_pos,
                                                                                polya_info.external_polyt_pos)
        split_exon_profile = self.split_exon_profile_constructor.construct_profile(sorted_blocks,
                                                                                   polya_info.external_polya_pos,
                                                                                   polya_info.external_polyt_pos)
        return CombinedReadProfiles(intron_profile, exon_profile, split_exon_profile,
                                    polya_info=polya_info, cage_hits=cage_hits)


############################################################################
# Copyright (c) 2022-2026 University of Helsinki
# Copyright (c) 2019-2022 Saint Petersburg State University
# # All Rights Reserved
# See file LICENSE for details.
############################################################################

import unittest
from src.common import junctions_from_blocks
from src.junction_comparator import *


class TestJunctionComparator(unittest.TestCase):
    def test_overlaps_at_least(self):
        junction1 = (100, 200)
        junction2 = (150, 250)
        self.assertTrue(overlaps_at_least(junction1, junction2, 10))
        
    def test_equal_ranges_with_delta(self):
        junction1 = (100, 200)
        junction2 = (102, 198)
        self.assertTrue(equal_ranges(junction1, junction2, delta=5))
        self.assertFalse(equal_ranges(junction1, junction2, delta=1))
        
    def test_junctions_from_blocks(self):
        blocks = [(100, 200), (300, 400), (500, 600)]
        junctions = junctions_from_blocks(blocks)
        self.assertEqual(junctions, [(201, 299), (401, 499)])
        
    def test_junctions_from_single_block(self):
        blocks = [(100, 200)]
        junctions = junctions_from_blocks(blocks)
        self.assertEqual(junctions, [])
